import re
import time
from copy import deepcopy
from typing import Tuple, List

from cobra import Reaction, Model, Metabolite

from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.model_seed.model_seed_compounds_database import ModelSeedCompoundsDB
from boimmgpy.service.model_mapper import ModelMapper
from boimmgpy.utilities import chemo_utilities, model_utilities, file_utilities
from boimmgpy.definitions import ROOT_DIR, REACTIONS_ANNOTATION_CONFIGS_PATH

PROGRESS_BAR = ROOT_DIR + "/service/logs/progress_bar.txt"
# PROGRESS_BAR = "/workdir/resultsWorker/progress_bar.txt"

def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='|', printEnd="\r"):
    """
  Call in a loop to create terminal progress bar
  @params:
      iteration   - Required  : current iteration (Int)
      total       - Required  : total iterations (Int)
      prefix      - Optional  : prefix string (Str)
      suffix      - Optional  : suffix string (Str)
      decimals    - Optional  : positive number of decimals in percent complete (Int)
      length      - Optional  : character length of bar (Int)
      fill        - Optional  : bar fill character (Str)
      printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
  """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="",flush=True)
    # Print New Line on Complete
    if iteration == total:
        print()

class Granulator:

    def __init__(self, model: Model, mapper: ModelMapper, database_format: str, id_converter: CompoundsIDConverter,
                 compoundsAnnotationConfigs: dict):

        """
        Class constructor

        :param Model model: Cobrapy model
        :param ModelMapper mapper: model mapper
        :param str database_format: ModelSEED, KEGG or BiGG
        :param CompoundsIDConverter id_converter:
        :param dict compoundsAnnotationConfigs: compounds annotation configurations
        """

        self.__compoundsAnnotationConfigs = compoundsAnnotationConfigs
        self.__model = model
        self.__compoundsIdConverter = id_converter
        self.__modelseedCompoundsDb = ModelSeedCompoundsDB()
        self.__database_format = database_format
        self.mapper = mapper

        self.define_instance_variables()

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, value):
        self.__model = value

    @property
    def mapper(self):
        return self.__mapper

    @mapper.setter
    def mapper(self, value):
        self.__mapper = value

    @property
    def virtual_model_mapper(self):
        return self.__virtual_model_mapper

    @virtual_model_mapper.setter
    def virtual_model_mapper(self, value):
        self.__virtual_model_mapper = value

    def define_instance_variables(self):
        """
        Instance variables definition

        :return:
        """

        self.__compounds_ontology  = CompoundsDBAccessor()

        self.__added_reactions_and_respective_target = {}

        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)

        self.metabolite_report_material = {}
        self.reaction_report_material = {}
        self.__set_components()

    def set_virtual_model(self,virtual_model, virtual_model_mapper, virtual_compounds_revisor):

        self.__virtual_model = virtual_model
        self.__virtual_compounds_revisor = virtual_compounds_revisor
        self.__virtual_model_mapper = virtual_model_mapper

    def write_in_progress_bar(self, message, value):
        with open(PROGRESS_BAR,"w") as file:
            file.write(message + ":" + str(value))

    def get_progress_bar_state(self):
        with open(PROGRESS_BAR,"r") as file:
            line = file.read()
            line = line.strip()
            state = line.split(":")[1]
            return float(state)

    def calculate_division_for_progress_bar(self,state,total,total_processes_left):
        left = 100 - state
        for_each_part = left / total_processes_left
        for_each_iteration = for_each_part / total

        return for_each_iteration

    def granulate(self, target_generic_ontology_id, components, same_components, progress_bar_processes_left):
        """
        Operation to granulate a given compound:

            1st: Build up structurally defined compounds using the information on the components list. It builds up
            compounds by mixing :param components if :param same_components is False, otherwise it builds up compounds
            without a mix of :param components.

            2nd: Calls a function to generate the whole biosynthetic pathway of all the built up compounds.

        :param int target_generic_ontology_id: generic compound database identifier
        :param list<int> components: list of BOIMMG identifiers of all the requested components
        :param bool same_components:
        :param int progress_bar_processes_left:
        :return:
        """


        if same_components:
            targets_to_replace = \
                self.__compounds_ontology.get_compounds_with_only_one_component(target_generic_ontology_id, components)

        else:
            targets_to_replace = \
                self.__compounds_ontology.get_compounds_with_specific_parent_within_set_of_components(
                    target_generic_ontology_id,
                    components)

        targets_to_replace = self.filter_targets_to_replace(targets_to_replace,target_generic_ontology_id)

        state = self.get_progress_bar_state()
        for_each_iteration = self.calculate_division_for_progress_bar(
                                        state,len(targets_to_replace),progress_bar_processes_left)

        for target in targets_to_replace:

            self.write_in_progress_bar("generating network for BMGC"+ str(target),state)
            self.handle_target_in_virtual_model(target, target_generic_ontology_id)
            state += for_each_iteration


    def handle_target_in_virtual_model(self, target: int, parent: int):
        """
        Method to granulate a given lipid target in the virtual model.

            1st: Processes the target biosynthesis pathway;
            2nd: Balance the granulated reactions in the previously processed biosynthesis pathway;
            3rd: Adds them to the model

        :param int target: lipid target BOIMMG identifier (structurally defined)
        :param int parent: lipid parent BOIMMG identifier (generic lipid)
        :return:
        """

        granulated_reactions = self.__process_biosynthesis_pathway(target, parent)


        granulated_reactions = self.__virtual_compounds_revisor.balance_reactions(
            granulated_reactions)

        reactions_to_add = deepcopy(granulated_reactions)

        self.__add_new_reactions_to_model(reactions_to_add)


    def identify_biosynthesis_pathway(self, target_generic_ontology_id: int) -> list:
        """
        Identifies the biosynthesis pathway of a given generic lipid.

        :param int target_generic_ontology_id:
        :return list<Reaction>: list of reactions
        """

        reactions_to_delete = self.get_target_associated_network(target_generic_ontology_id)

        return reactions_to_delete


    def __process_biosynthesis_pathway(self, target: int, parent: int) -> list:
        """
        Granulates the biosynthesis pathway of a given target lipid. It uses a "processing" stack to process each reaction
        and each new structurally defined lipid. When the stack is empty it is assumed the pathway to be fully processed.

        :param int target: lipid target BOIMMG identifier (structurally defined)
        :param int parent: lipid parent BOIMMG identifier (generic lipid)
        :return list<Reaction>: list of granulated reactions
        """

        processed_compounds = []
        new_compounds = [(target, parent)]
        granulated_reactions = []

        while new_compounds:
            new_compound_and_parent = new_compounds.pop()

            if new_compound_and_parent[0] not in processed_compounds:


                new_reactions, added_compounds = self.__process_new_compound(new_compound_and_parent)

                processed_compounds.append(new_compound_and_parent[0])

                self.__add_new_reactions_to_virtual_model(new_reactions)

                granulated_reactions.extend(new_reactions)

                new_compounds.extend(added_compounds)

        return granulated_reactions


    def __process_new_compound(self, new_compound_and_parent: tuple):
        """
        Processes a new compound

        :param tuple new_compound_and_parent: (structurally defined compound, its generic parent)
        :return:
        """

        new_compound = new_compound_and_parent[0]

        parent = new_compound_and_parent[1]
        model_boimmg_parents = self.virtual_model_mapper.check_if_boimmg_metabolite_in_model(parent)

        if not model_boimmg_parents:
            exit(2)
            raise Exception("please introduce a mapped biosynthetic target")

        parent_and_precursors = self.get_targets_to_replace_by_type(new_compound)

        new_reactions = []

        added_compounds = []

        processed = []
        for model_parent in model_boimmg_parents:

            if model_parent not in processed:
                cobra_model_parent = self.__virtual_model.metabolites.get_by_id(model_parent)
                reactions_to_process = self.get_reactions_by_compound(cobra_model_parent)

                for reaction_to_process in reactions_to_process:

                    if reaction_to_process.id not in self.__added_reactions_and_respective_target.keys() or \
                            new_compound not in self.__added_reactions_and_respective_target[reaction_to_process.id]:

                        go = True
                        compounds_ids = [comp.id for comp in reaction_to_process.products]

                        if go and model_parent in compounds_ids:
                            new_reaction, added = self._process_reaction(reaction_to_process, new_compound,
                                                                         cobra_model_parent, parent_and_precursors)


                            if reaction_to_process.id in self.reaction_report_material:
                                self.reaction_report_material[reaction_to_process.id].append(new_reaction.id)
                            else:
                                self.reaction_report_material[reaction_to_process.id] = [new_reaction.id]

                            new_reactions.append(new_reaction)

                            for compound in added:
                                if compound not in added_compounds:
                                    added_compounds.append(compound)

                            if reaction_to_process.id not in self.__added_reactions_and_respective_target.keys():
                                self.__added_reactions_and_respective_target[reaction_to_process.id] = [new_compound]

                            else:
                                self.__added_reactions_and_respective_target[reaction_to_process.id].append(new_compound)

                processed.append(model_parent)

        return new_reactions, added_compounds


    def get_reactions_by_compound(self, compound):


        res = []
        for reaction in self.__virtual_model.reactions:
            metabolites_ids = [m.id for m in reaction.metabolites]

            if compound.id in metabolites_ids:
                res.append(reaction)

        return res


    def _process_reaction(self, reaction_to_process: Reaction, new_compound: int, model_parent: Metabolite,
                          parent_and_precursors: dict) -> Tuple[Reaction, List[Tuple[Metabolite, int]]]:
        """
        Granulates the :param reaction_to_process

        :param Reaction reaction_to_process: reaction in model to be granulated
        :param int new_compound: structurally defined lipid BOIMMG identifier
        :param Metabolite model_parent: model generic lipid
        :param dict parent_and_precursors: {parent: [children]} -> children of the generic precursors
        :return tuple: the granulated reaction in the cobrapy Reaction object and list of tuples: (added compound, BOIMMG ID)
        """

        new_reaction = deepcopy(reaction_to_process)

        new_reaction_products = list(new_reaction.products)

        new_reaction, added = self.__process_reaction(new_reaction, new_compound, model_parent, parent_and_precursors,
                                                          new_reaction_products)

        reaction_id = self.change_boimmg_reaction_id(new_reaction)

        new_reaction.id = reaction_id

        return new_reaction, added

    def change_boimmg_reaction_id(self, reaction: Reaction) -> str:

        """
        Changes the reaction ID into a cannonical BOIMMG reaction identifier

        :param Reaction reaction:
        :return str: new identifier
        """

        new_id = self.__reactionsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]

        metabolites = reaction.metabolites

        metabolites = sorted([metabolite.id for metabolite in metabolites])
        new_id += "_".join(metabolites)

        if len(new_id) > 250:
            new_id = new_id[:230]

        reaction.id = new_id

        return reaction.id



    def __process_reaction(self, new_reaction: Reaction, new_compound: int, model_parent: Metabolite,
                           parent_and_precursors: dict,
                           new_reaction_metabolites: List[Metabolite]) -> Tuple[Reaction, List[Tuple[Metabolite, int]]]:

        """
        Granulates a given reaction. Replaces the generic metabolite by its structurally defined child and corrects the
        precursors.

        :param Reaction new_reaction: the new reaction
        :param int new_compound: the reaction structurally defined product
        :param Metabolite model_parent: the model parent of the structurally defined product
        :param dict parent_and_precursors: dictionary with lipid parents and their children
        :param List[Metabolite] new_reaction_metabolites: list of possibly new metabolites
        :return Tuple[Reaction, List[Tuple[Metabolite, int]]]: returns a tuple with the granulated model reaction
        and the new added compounds that will be processed further.
        """

        i = 0
        found = False

        while not found and i < len(new_reaction_metabolites):
            if new_reaction_metabolites[i].id == model_parent.id:
                new_model_parent = new_reaction_metabolites[i]
                found = True

            i += 1

        coef = new_reaction.get_coefficient(model_parent.id)

        new_metabolite = self.add_boimmg_metabolites_to_reaction(new_compound, coef, new_reaction, model_parent.compartment)

        if model_parent.id in self.metabolite_report_material:
            self.metabolite_report_material[model_parent.id].append(new_metabolite.id)
        else:
            self.metabolite_report_material[model_parent.id] = [new_metabolite.id]

        met_to_subtract = {}

        met_to_subtract[new_model_parent] = coef
        new_reaction.subtract_metabolites(met_to_subtract)

        added = self._correct_precursors(new_reaction, new_compound, parent_and_precursors)

        return new_reaction, added


    def _correct_precursors(self, new_reaction: Reaction, new_compound:int, parent_and_precursors:dict) -> \
            List[Tuple[Metabolite, int]]:

        """
        Correct the precursors in the new reaction.

        :param Reaction new_reaction: the new reaction
        :param int new_compound: the new structurally defined lipid species BOIMMG identifier
        :param dict parent_and_precursors: dictionary with lipid parents and their children
        :return List[Tuple[Metabolite, int]]: returns a tuple with the granulated model reaction
        and the new added compounds that will be processed further.
        """

        new_compound_components = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(new_compound,
                                                                                                "component_of")

        reactants = list(new_reaction.reactants)
        added = self.__correct_precursors(new_reaction, parent_and_precursors, reactants,
                                              new_compound_components)

        return added


    def __correct_precursors(self, new_reaction: Reaction, parent_and_precursors: dict,
                             metabolites: List[Metabolite], new_compound_components: List[int]) -> \
                            List[Tuple[Metabolite, int]]:

        """
        Correct the precursors in the new reaction.

        :param Reaction new_reaction: the new reaction
        :param dict parent_and_precursors: dictionary with lipid parents and their children
        :param List[Metabolite] metabolites: number of reactants in the new reaction to be granulated
        :param List[int] new_compound_components: components of the new product in the reaction
        :return:
        """

        added = []
        while metabolites:

            reactant = metabolites.pop()

            boimmg_id = self.virtual_model_mapper.get_boimmg_id_from_model_compound_id(reactant.id)
            if boimmg_id:

                coef = new_reaction.metabolites[reactant]
                if boimmg_id in parent_and_precursors:

                    precursors = parent_and_precursors[boimmg_id]
                    for precursor in precursors:
                        self.add_boimmg_metabolites_to_reaction(precursor, coef, new_reaction, reactant.compartment)

                        added.append((precursor, boimmg_id))

                    met_to_subtract = {}

                    met_to_subtract[reactant] = coef
                    new_reaction.subtract_metabolites(met_to_subtract)



                else:
                    to_replace = self.__compounds_ontology.get_compounds_with_specific_parent_set_of_components(
                        boimmg_id,
                        new_compound_components)

                    if to_replace:
                        to_replace = self.filter_targets_to_replace(to_replace,boimmg_id)

                        self.add_boimmg_metabolites_to_reaction(to_replace[0], coef, new_reaction, reactant.compartment)

                        met_to_subtract = {}

                        met_to_subtract[reactant] = coef
                        new_reaction.subtract_metabolites(met_to_subtract)

                        added.append((to_replace[0], boimmg_id))
        return added


    def add_boimmg_metabolites_to_reaction(self, boimmg_id: int, coef: int, reaction: Reaction, compartment: str):
        """
        Add a metabolite to a given reaction receiving BOIMMG identifier as input.

        :param int boimmg_id: BOIMMG identifier of the metabolite being inserted in the reaction
        :param int coef: stoichimetric coefficient in the reaction
        :param Reaction reaction: model reaction
        :param str compartment: compartment
        :return:
        """

        model_seed_ids = self.__compounds_ontology.get_all_model_seed_ids(boimmg_id)
        node = self.__compounds_ontology.get_node_by_ont_id(boimmg_id)

        aliases = node.aliases

        found = False
        i = 0
        while not found and i<len(model_seed_ids):

            aliases2 = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_ids[i])

            aliases.update(aliases2)

            model_child = self.mapper.check_if_boimmg_metabolite_in_model(boimmg_id,aliases)

            if model_child:
                found = True

            i+=1

        if not model_seed_ids:
            model_child = self.mapper.check_if_boimmg_metabolite_in_model(boimmg_id)


        if not model_child:

            new_model_compound = self.__create_boimmg_metabolite(boimmg_id, compartment)

            met_to_add = {}

            met_to_add[new_model_compound] = coef
            reaction.add_metabolites(met_to_add)


        else:
            found = False
            i = 0
            while not found and i < len(model_child):
                cobra_child_container = self.model.metabolites.get_by_id(model_child[i])
                if cobra_child_container.compartment == compartment:
                    found = True

                    if self.__virtual_model.metabolites.has_id(model_child[i]):
                        found_model_compound = cobra_child_container

                    else:
                        virtual_compound = deepcopy(cobra_child_container)
                        self.__virtual_model.add_metabolites([virtual_compound])
                        found_model_compound = virtual_compound

                i += 1

            if not found:
                new_model_compound = self.__create_boimmg_metabolite(boimmg_id, compartment)

                met_to_add = {}

                met_to_add[new_model_compound] = coef
                reaction.add_metabolites(met_to_add)

            else:
                met_to_add = {}

                new_model_compound = found_model_compound
                met_to_add[found_model_compound] = coef
                reaction.add_metabolites(met_to_add)

        return new_model_compound

    def __create_boimmg_metabolite(self, boimmg_id:int, compartment:str) -> Metabolite:

        """
        Generates a new model compound using the internal database information.

        :param int boimmg_id: BOIMMG identifier
        :param str compartment: compartment
        :return Metabolite: the new metabolite
        """

        new_compound_node = self.__compounds_ontology.get_node_by_ont_id(boimmg_id)

        model_seed_ids = self.__compounds_ontology.get_all_model_seed_ids(boimmg_id)

        for model_seed_id in model_seed_ids:

            translation = self.__compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id)

            definitive_model_seed_id = model_seed_id
            if self.__database_format == "BiGG" and \
                    ("BiGG" in translation.keys() or "BiGG1" in translation.keys()):

                definitive_model_seed_id = model_seed_id
                break


            elif self.__database_format == "KEGG" and \
                    "KEGG" in translation.keys():

                definitive_model_seed_id = model_seed_id
                break

        new_model_compound = model_utilities.generate_boimmg_metabolites_in_compartment(new_compound_node, compartment,
                                                                                       self.__compoundsIdConverter,
                                                                                       self.__compoundsAnnotationConfigs)
        self.add_metabolites_to_the_virtual_model([new_model_compound])

        return new_model_compound


    def add_metabolites_to_the_virtual_model(self, metabolites):
        self.__virtual_model.add_metabolites(metabolites)
        self.virtual_model_mapper.add_new_metabolites_to_maps(metabolites)


    @staticmethod
    def merge_dictionaries(dict1, dict2):
        for key in dict1:
            if key in dict2:
                dict1[key].extend(dict2[key])
        return dict1


    def get_target_associated_network(self, target_generic_ontology_id: int) -> List[Reaction]:
        """
        Identifies the target's biosynthesis network in the model

        :param int target_generic_ontology_id:
        :return List[Reaction]: list of reactions in the network
        """

        res = []
        res_ids = []

        predecessors = \
            self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(target_generic_ontology_id, "precursor_of")

        predecessors_and_target = predecessors.copy()
        predecessors_and_target.append(target_generic_ontology_id)

        network_metabolites_in_model = []

        for token in predecessors_and_target:
            metabolites_in_model = self.__check_if_compound_exists_in_model_by_ontology_id(token)
            network_metabolites_in_model.extend(metabolites_in_model)

        network_metabolites_ids = [m.id for m in network_metabolites_in_model]

        for metabolites_in_model in network_metabolites_in_model:
            reactions = metabolites_in_model.reactions
            for reaction in reactions:

                if reaction.id not in res_ids:

                    reactants = reaction.reactants
                    products = reaction.products

                    reactant_found = False
                    product_found = False
                    i = 0
                    while not reactant_found and i < len(reactants):
                        reactant = reactants[i]
                        reactant_id = reactant.id
                        if reactant_id in network_metabolites_ids:
                            reactant_found = True

                        i += 1

                    j = 0
                    while not product_found and j < len(products):
                        product = products[j]
                        product_id = product.id
                        if product_id in network_metabolites_ids:
                            product_found = True
                        j += 1

                    if product_found and reactant_found:
                        res.append(reaction)
                        res_ids.append(reaction.id)


        return res


    def __check_if_compound_exists_in_model_by_ontology_id(self, ontology_id: int) -> List[Metabolite]:
        """
        This method will check whether a BOIMMG compound exists in the model.

        :param int ontology_id: BOIMMG id
        :return List[Metabolite]: compounds in model
        """
        res = []

        for model_id, boimmg_id in self.mapper.boimmg_db_model_map.items():
            if boimmg_id == ontology_id:
                model_compound_container = self.model.metabolites.get_by_id(model_id)
                res.append(model_compound_container)

        return res

    def __add_new_reactions_to_virtual_model(self,new_reactions):
        self.virtual_model_mapper.add_new_reactions_to_model(new_reactions)
        self.__virtual_model.add_reactions(new_reactions)

    def __add_new_reactions_to_model(self,new_reactions):
        self.mapper.add_new_reactions_to_model(new_reactions)
        self.model.add_reactions(new_reactions)

    def get_targets_to_replace_by_type(self, target: int) -> dict:
        """
        Identifies targets components, precursors and successors in the biosynthesis network. Returns a dictionary
        with the precursors' generic parent as keys and a list of structurally defined precursors as values.

        :param int target: BOIMMG identifier
        :return dict: dictionary with the precursors' generic parent as keys and a list
        of structurally defined precursors as values
        """

        res={}
        precursors = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(target,"precursor_of")

        successors = self.__compounds_ontology.get_successors_by_ont_id_rel_type(target,"precursor_of")

        components = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(target,"component_of")

        for component in components:
            if component in self.fatty_acids:
                if self.generic_fatty_acid in res:
                    res.get(self.generic_fatty_acid).append(component)
                else:
                    res[self.generic_fatty_acid] = [component]

        for precursor in precursors:
            parent = self.__compounds_ontology.get_successors_by_ont_id_rel_type(precursor,"is_a")

            if parent[0] in res:
                res.get(parent[0]).append(precursor)
            else:
                res[parent[0]] = [precursor]

        for successor in successors:
            parent = self.__compounds_ontology.get_successors_by_ont_id_rel_type(successor,"is_a")

            if parent[0] in res:
                res.get(parent[0]).append(successor)
            else:
                res[parent[0]] = [successor]


        return res

    # def find_components_reactions_and_generalize(self):
    #
    #     candidate_reactions = []
    #
    #     res = {}
    #
    #     generic_compounds = []
    #     pseudo_components = []
    #     components = []
    #     components_boimmg_id = []
    #
    #     for compound in self.mapper.boimmg_db_model_map:
    #
    #         boimmg_id =self.mapper.boimmg_db_model_map[compound]
    #         if boimmg_id in self.__components:
    #             model_compound = self.model.metabolites.get_by_id(compound)
    #             pseudo_components.append(compound)
    #             candidate_reactions.extend(list(model_compound.reactions))
    #
    #             if boimmg_id in self.fatty_acids:
    #                 components.append(compound)
    #
    #         node = self.__compounds_ontology.get_node_by_ont_id(self.mapper.boimmg_db_model_map[compound])
    #
    #         if node.generic:
    #             generic_compounds.append(compound)
    #
    #     for reaction in candidate_reactions:
    #
    #         regex = re.compile("[Bb][Ii][Oo][Mm][Aa][Ss][Ss]")
    #
    #         if reaction.annotation != "SBO:0000629" and not regex.match(reaction.name):
    #
    #             reactants = reaction.reactants
    #             products = reaction.products
    #
    #             reactant_found = False
    #             product_found = False
    #
    #             pseudo_component_in_reactants = False
    #
    #             isComponent = False
    #
    #             for reactant in reactants:
    #                 reactant_id = reactant.id
    #
    #                 if reactant_id in generic_compounds:
    #                     reactant_found = True
    #
    #                 elif reactant_id in pseudo_components:
    #                     pseudo_component_in_reactants = True
    #
    #                     if reactant_id in components:
    #                         isComponent = True
    #
    #             for product in products:
    #                 product_id = product.id
    #                 if product_id in generic_compounds:
    #                     product_found = True
    #
    #                 elif product_id in components:
    #                     isComponent = True
    #
    #             if reactant_found or product_found:
    #
    #                 new_reaction, subtracted = \
    #                     self.remove_components_from_reaction(reaction,pseudo_components,pseudo_component_in_reactants,isComponent)
    #
    #                 for sub in subtracted:
    #                     if sub not in components_boimmg_id:
    #                         components_boimmg_id.append(sub)
    #
    #
    #                 res[reaction.id] = (new_reaction,isComponent,pseudo_component_in_reactants,subtracted)
    #
    #     return res,components_boimmg_id

    # def remove_components_from_reaction(self,reaction,pseudo_components,pseudo_component_in_reactants,isComponent):
    #
    #     new_reaction = deepcopy(reaction)
    #
    #     component_parent = None
    #     if pseudo_component_in_reactants:
    #
    #         reactants = list(new_reaction.reactants)
    #
    #         to_subtract = {}
    #
    #         while reactants:
    #             reactant = reactants.pop()
    #             if reactant.id in self.mapper.boimmg_db_model_map.keys():
    #                 if reactant.id in pseudo_components:
    #
    #                     if not component_parent:
    #                         child = self.mapper.get_boimmg_id_from_model_compound_id(reactant.id)
    #                         parent = self.__compounds_ontology.get_successors_by_ont_id_rel_type(child, "is_a")
    #                         component_parent = parent[0]
    #
    #                     coef = new_reaction.metabolites[reactant]
    #                     to_subtract[reactant] = coef
    #
    #         new_reaction.subtract_metabolites(to_subtract)
    #
    #         components = []
    #         for subtracted in to_subtract:
    #             if isComponent:
    #                 boimmg_id = self.mapper.get_boimmg_id_from_model_compound_id(subtracted.id)
    #                 components.append(boimmg_id)
    #
    #             else:
    #                 boimmg_id = self.mapper.get_boimmg_id_from_model_compound_id(subtracted.id)
    #                 new_components = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(boimmg_id,
    #                                                                                                "component_of")
    #                 components.extend(new_components)
    #
    #         if isComponent:
    #             self.add_boimmg_metabolites_to_reaction(self.generic_fatty_acid, -1, new_reaction, reactant.compartment)
    #
    #         else:
    #
    #             self.add_boimmg_metabolites_to_reaction(component_parent,-1,new_reaction,reactant.compartment)
    #
    #         return new_reaction, components
    #
    #     else:
    #         products = list(new_reaction.products)
    #
    #         to_subtract = {}
    #
    #         while products:
    #             product = products.pop()
    #
    #             if product.id in self.mapper.boimmg_db_model_map.keys():
    #                 if product.id in pseudo_components:
    #
    #                     if not component_parent:
    #
    #                         child = self.mapper.get_boimmg_id_from_model_compound_id(product.id)
    #                         parent = self.__compounds_ontology.get_successors_by_ont_id_rel_type(child, "is_a")
    #
    #                         component_parent = parent[0]
    #
    #                     coef = new_reaction.metabolites[product]
    #                     to_subtract[product] = coef
    #
    #         if isComponent:
    #             self.add_boimmg_metabolites_to_reaction(self.generic_fatty_acid, 1, new_reaction, product.compartment)
    #
    #         else:
    #
    #             self.add_boimmg_metabolites_to_reaction(component_parent, 1, reaction, new_reaction.compartment)
    #
    #         new_reaction.subtract_metabolites(to_subtract)
    #
    #         components = []
    #         for subtracted in to_subtract:
    #             if isComponent:
    #                 boimmg_id = self.mapper.get_boimmg_id_from_model_compound_id(subtracted.id)
    #                 components.append(boimmg_id)
    #
    #             else:
    #                 boimmg_id = self.mapper.get_boimmg_id_from_model_compound_id(subtracted.id)
    #                 new_components = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(boimmg_id,
    #                                                                                                "component_of")
    #                 components.extend(new_components)
    #
    #
    #         return new_reaction, components


    # def __get_children_in_model(self,product):
    #
    #     children_in_model = {}
    #
    #     parent_boimmg_id = self.virtual_model_mapper.get_boimmg_id_from_model_compound_id(product.id)
    #
    #     if parent_boimmg_id:
    #         children = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(parent_boimmg_id,"is_a")
    #
    #         for mapped_compound in self.virtual_model_mapper.boimmg_db_model_map:
    #             eventual_child = self.virtual_model_mapper.boimmg_db_model_map[mapped_compound]
    #
    #             if eventual_child in children:
    #
    #                 children_in_model[eventual_child] = self.virtual_model_mapper.boimmg_db_model_map[mapped_compound]
    #
    #     return children_in_model

    # def check_if_decomposition_reaction(self,reaction:Reaction):
    #
    #     reactants = reaction.reactants
    #     products = reaction.products
    #
    #     inReactants = self.__check_component_in_reaction_side(reactants)
    #
    #     inProducts = self.__check_component_in_reaction_side(products)
    #
    #     if inReactants or inProducts:
    #         return True,inReactants
    #
    #     else:
    #         return False, inReactants



    # def __check_component_in_reaction_side(self,metabolites):
    #     for metabolite in metabolites:
    #
    #         boimmg_id = self.virtual_model_mapper.get_boimmg_id_from_model_compound_id(metabolite.id)
    #
    #         if boimmg_id:
    #             if boimmg_id == self.generic_fatty_acid or boimmg_id == self.__generic_acps or \
    #                     boimmg_id == self.__generic_acyl_coa:
    #                 return True
    #
    #     else:
    #         return False

    def __set_components(self):

        self.fatty_acids = self.__compounds_ontology.get_leaves_from_ont_id(527)
        self.generic_fatty_acid = 527

        # self.__acyl_coas = self.__compounds_ontology.get_leaves_from_ont_id(43)
        # self.__generic_acyl_coa = 43
        #
        # self.__acyl_acps = self.__compounds_ontology.get_leaves_from_ont_id(704832)
        # self.__generic_acps = 704832

        # self.__components = self.fatty_acids + self.__acyl_acps + self.__acyl_coas
        # self.__components = self.fatty_acids

    # def solve_components_reactions(self,same_components=False,components_reactions = {}, components = []):
    #
    #     if not components_reactions and not components:
    #         components_reactions,components = self.find_components_reactions_and_generalize()
    #
    #     new_reactions = [components_reactions[reaction][0] for reaction in components_reactions.keys()]
    #
    #     self.__add_new_reactions_to_virtual_model(new_reactions)
    #
    #     self.__solve_components_reactions(components_reactions,components,same_components)

    def __solve_components_reactions(self,components_reactions,components,same_components):


        for reaction in components_reactions:

            self.__correct_component_reaction(components_reactions[reaction],components,same_components)

    # def __correct_component_reaction(self,reaction_features,components,same_components):
    #
    #     new_reaction = reaction_features[0]
    #     isReactant = reaction_features[2]
    #     deleted = reaction_features[3]
    #
    #     reactants = new_reaction.reactants
    #     products = new_reaction.products
    #
    #     # r_groups_reactants,r_groups_products = self.get_r_groups_from_reaction_compounds(reactants,products)
    #
    #     if isReactant:
    #
    #         boimmg_compound = self.find_boimmg_compound(products)
    #
    #         if boimmg_compound:
    #
    #             self.granulate(boimmg_compound, components, same_components)
    #
    #     else:
    #         boimmg_compound = self.find_boimmg_compound(reactants)
    #
    #         if boimmg_compound:
    #             self.granulate(boimmg_compound, components, same_components,decomposition=True)


    def find_boimmg_compound(self,compounds):
        """
        It finds only one boimmg compound within a list of model compounds
        :param compounds:
        :return:
        """

        for compound in compounds:

            boimmg_id = self.mapper.get_boimmg_id_from_model_compound_id(compound.id)

            if boimmg_id:
                return boimmg_id

        return None


    # def get_r_groups_from_reaction_compounds(self,reactants,products):
    #
    #     r_groups_reactants = 0
    #     r_groups_products = 0
    #
    #     for reactant in reactants:
    #         r_groups_reactants += self.get_r_groups_from_compound(reactant)
    #
    #     for product in products:
    #         r_groups_products += self.get_r_groups_from_compound(product)
    #
    #     return r_groups_reactants,r_groups_products

    # def get_r_groups_from_compound(self,model_compound):
    #
    #     boimmg_id = self.virtual_model_mapper.get_boimmg_id_from_model_compound_id(model_compound.id)
    #     if boimmg_id:
    #         node = self.__compounds_ontology.get_node_by_ont_id(boimmg_id)
    #         smiles = node.smiles
    #
    #         if smiles:
    #             return smiles.count("*")
    #
    #         return 0
    #     return 0

    def filter_targets_to_replace(self,targets: List[int],parent: int) -> List[int]:
        """
        Filter the predicted precursors by side chains.

        :param List[int] targets:
        :param int parent:
        :return List[int]: filtered list
        """

        res =  []

        for target in targets:
            components = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(target,"component_of")

            node = self.__compounds_ontology.get_node_by_ont_id(target)

            child_smiles = node.smiles

            parent_node = self.__compounds_ontology.get_node_by_ont_id(parent)

            parent_smiles = parent_node.smiles

            molecules = chemo_utilities.get_side_chains(parent_smiles, child_smiles)

            if molecules:
                new_molecules = []
                new_molecules_n = 0
                for molecule in molecules:

                    if molecule not in new_molecules:
                        new_molecules_n+=1
                        new_molecules.append(molecule)

                if new_molecules_n == len(components):
                    res.append(target)

        return res


from boimmgpy.service.network_handlers.graph import Graph


class MetabolicNetwork(Graph):

    def __init__(self, network_type = "metabolite-reaction", split_rev=False):
        Graph.__init__(self,{})
        self.net_type=network_type
        self.node_types={}
        if network_type == "metabolite-reaction":
            self.node_types["metabolite"]=[]
            self.node_types["reaction"]=[]
        elif network_type == "metabolite-metabolite":
            self.node_types["metabolite"] = []
        self.split_rev=split_rev

    def convert_metabolite_net(self,gmr):
        for m in gmr.node_types["metabolite"]:
            self.add_vertex(m)
            sucs = gmr.get_successors(m)
            for s in sucs:
                sucs_r = gmr.get_successors(s)
                for s2 in sucs_r:
                    if m != s2:
                        self.add_edge(m, s2)

    def convert_reaction_graph(self,gmr):
        for r in gmr.node_types["reaction"]:
            self.add_vertex(r)
            sucs=gmr.get_successors(r)
            for s in sucs:
                sucs_r=gmr.get_successors(s)
                for s2 in sucs_r:
                    if r != s2:
                        self.add_edge(r,s2)

    def add_vertex_type(self, v , nodetype):
        self.add_vertex(v)
        self.node_types[nodetype].append(v)

    def get_nodes_type(self,node_type):
        if node_type in self.node_types:
            return self.node_types[node_type]
        else:
            return None

    def get_pathway(self):
        longer_path = []
        for node in self.get_nodes():
            path = self.reachable_bfs(node)
            if len(path)>len(longer_path):
                longer_path=path
        return longer_path

    def get_longest_pathway(self,node):
        longer_path = []
        path = self.reachable_bfs(node)
        if len(path)>len(longer_path):
            longer_path=path
        return longer_path

    def get_complete_pathway(self):
        pathway = self.get_pathway()
        node = pathway[0]
        return self.reachable_bfs(node)

    def get_all_targets(self):
        res = []
        for token in self.graph:
            successors = self.get_successors(token)
            predecessors = self.get_predecessors(token)
            if not successors:
                res.append(token)

            elif successors == predecessors:
                res.append(token)

        return res

    def prune_redundant_cycles(self):
        old_graph = self.graph.copy()

        if len(old_graph) > 1:
            for token in old_graph:
                successors = self.get_successors(token)
                predecessors = self.get_predecessors(token)

                if successors == predecessors:
                    self.delete_node(token)



    def get_starting_point(self):
        res = []
        for token in self.graph:
            predecessors = self.get_predecessors(token)
            if not predecessors:
                res.append(token)
        return res

    def get_all_target_pathways(self):
        starting_points = self.get_starting_point()
        targets = self.get_all_targets()

        res = []
        if len(self.graph)>1:
            for point in starting_points:
                for target in targets:
                    sh_path = self.shortest_path(point,target)

                    if sh_path:
                        res.append(sh_path)
        else:
            res = [starting_points]
        return res








import cobra

from boimmgpy import definitions


def map_e_coli_granulated_model():
    #################### pe ##################33
    # model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iAF1260b.xml")
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iML1515.xml")
    # model = cobra.io.read_sbml_model("C:/Users/Joao/Desktop/thesis_jcapela/BOIMMGpy/models/iAF1260b.xml")

    for c in ["c", "p"]:
        pe181 = model.metabolites.get_by_id("pe181_" + c)
        pe181.annotation["seed.compound"] = "cpd25142"
        pe181.annotation["boimmg.compound"] = "C_BOIMMG_8715"

        pe161 = model.metabolites.get_by_id("pe161_" + c)
        pe161.annotation["seed.compound"] = "cpd31682"
        pe161.annotation["boimmg.compound"] = "C_BOIMMG_12583"

        pe16 = model.metabolites.get_by_id("pe160_" + c)
        pe16.annotation["boimmg.compound"] = "C_BOIMMG_12595"

        pe14 = model.metabolites.get_by_id("pe140_" + c)
        pe14.annotation["seed.compound"] = "cpd33856"
        pe14.annotation["boimmg.compound"] = "C_BOIMMG_12604"

        pe141 = model.metabolites.get_by_id("pe141_" + c)
        pe141.annotation["boimmg.compound"] = "C_BOIMMG_12585"

        ############### pg ################

        pe141 = model.metabolites.get_by_id("pg141_" + c)
        pe141.annotation["boimmg.compound"] = "C_BOIMMG_7456"

        pe181 = model.metabolites.get_by_id("pg181_" + c)
        pe181.annotation["boimmg.compound"] = "C_BOIMMG_7408"

        pe141 = model.metabolites.get_by_id("pg140_" + c)
        pe141.annotation["seed.compound"] = "cpd35089"
        pe141.annotation["boimmg.compound"] = "C_BOIMMG_7474"

        pe160 = model.metabolites.get_by_id("pg160_" + c)
        pe160.annotation["boimmg.compound"] = "C_BOIMMG_427"

        pe161 = model.metabolites.get_by_id("pg161_" + c)
        pe161.annotation["boimmg.compound"] = "C_BOIMMG_7454"

        ############### ps ################

        if c == "c":


            ps160 = model.metabolites.get_by_id("ps160_" + c)
            ps160.annotation["boimmg.compound"] = "C_BOIMMG_729826"

            ps161 = model.metabolites.get_by_id("ps161_" + c)
            ps161.annotation["boimmg.compound"] = "C_BOIMMG_729871"

            ps140 = model.metabolites.get_by_id("ps140_" + c)
            ps140.annotation["boimmg.compound"] = "C_BOIMMG_729819"

            ps141 = model.metabolites.get_by_id("ps141_" + c)
            ps141.annotation["boimmg.compound"] = "C_BOIMMG_729870"

            ps180 = model.metabolites.get_by_id("ps180_" + c)
            ps180.annotation["boimmg.compound"] = "C_BOIMMG_729833"

            ps181 = model.metabolites.get_by_id("ps181_" + c)
            ps181.annotation["boimmg.compound"] = "C_BOIMMG_729912"

        ############## pgp ##############

        pgp161 = model.metabolites.get_by_id("pgp161_" + c)
        pgp161.annotation["boimmg.compound"] = "C_BOIMMG_296512"

        pgp140 = model.metabolites.get_by_id("pgp140_" + c)
        pgp140.annotation["boimmg.compound"] = "C_BOIMMG_296532"

        pgp141 = model.metabolites.get_by_id("pgp141_" + c)
        pgp141.annotation["boimmg.compound"] = "C_BOIMMG_296513"

        pgp160 = model.metabolites.get_by_id("pgp160_" + c)
        pgp160.annotation["boimmg.compound"] = "C_BOIMMG_296523"

        pgp181 = model.metabolites.get_by_id("pgp181_" + c)
        pgp181.annotation["boimmg.compound"] = "C_BOIMMG_293736"

        ################### cdpdag ######################3

        if c == "c":
            cdp141 = model.metabolites.get_by_id("cdpdtdec7eg_" + c)
            cdp141.annotation["boimmg.compound"] = "C_BOIMMG_574506"

            cdp181 = model.metabolites.get_by_id("cdpdodec11eg_" + c)
            cdp181.annotation["boimmg.compound"] = "C_BOIMMG_540741"

            # cdp180 = model.metabolites.get_by_id("cdpdodecg_" + c)
            # cdp180.annotation["boimmg.compound"] = "C_BOIMMG_540741"

            cdp160 = model.metabolites.get_by_id("cdpdhdecg_" + c)
            cdp160.annotation["seed.compound"] = "cpd23593"
            cdp160.annotation["boimmg.compound"] = "C_BOIMMG_239898"

            cdp140 = model.metabolites.get_by_id("cdpdtdecg_" + c)
            cdp140.annotation["seed.compound"] = "cpd32277"
            cdp140.annotation["boimmg.compound"] = "C_BOIMMG_574421"

            cdp161 = model.metabolites.get_by_id("cdpdhdec9eg_" + c)
            cdp161.annotation["boimmg.compound"] = "C_BOIMMG_524486"

        ############# PA ###############

        pa141 = model.metabolites.get_by_id("pa140_" + c)
        pa141.annotation["boimmg.compound"] = "C_BOIMMG_38435"

        pa141 = model.metabolites.get_by_id("pa141_" + c)
        pa141.annotation["boimmg.compound"] = "C_BOIMMG_38417"

        pa160 = model.metabolites.get_by_id("pa160_" + c)
        pa160.annotation["boimmg.compound"] = "C_BOIMMG_38416"

        pa161 = model.metabolites.get_by_id("pa161_" + c)
        pa161.annotation["boimmg.compound"] = "C_BOIMMG_423"

        pa180 = model.metabolites.get_by_id("pa181_" + c)
        pa180.annotation["seed.compound"] = "cpd33760"
        pa180.annotation["boimmg.compound"] = "C_BOIMMG_38371"

        ##############3 12dgr #################3

        dgr140 = model.metabolites.get_by_id("12dgr140_" + c)
        dgr140.annotation["seed.compound"] = "cpd16468"
        dgr140.annotation["boimmg.compound"] = "C_BOIMMG_53297"

        dgr141 = model.metabolites.get_by_id("12dgr141_" + c)
        dgr141.annotation["boimmg.compound"] = "C_BOIMMG_53244"

        dgr160 = model.metabolites.get_by_id("12dgr160_" + c)
        dgr160.annotation["seed.compound"] = "cpd26605"
        dgr160.annotation["boimmg.compound"] = "C_BOIMMG_61409"

        dgr161 = model.metabolites.get_by_id("12dgr161_" + c)
        dgr161.annotation["boimmg.compound"] = "C_BOIMMG_61238"

        dgr180 = model.metabolites.get_by_id("12dgr180_" + c)
        dgr180.annotation["boimmg.compound"] = "C_BOIMMG_70311"

        dgr181 = model.metabolites.get_by_id("12dgr181_" + c)
        dgr181.annotation["boimmg.compound"] = "C_BOIMMG_70100"

        ############## agpg ##############

        if c == "p":
            clpn161 = model.metabolites.get_by_id("clpn161_" + c)
            clpn161.annotation["boimmg.compound"] = "C_BOIMMG_341201"

            clpn160 = model.metabolites.get_by_id("clpn160_" + c)
            clpn160.annotation["boimmg.compound"] = "C_BOIMMG_322789"

            clpn181 = model.metabolites.get_by_id("clpn181_" + c)
            clpn181.annotation["boimmg.compound"] = "C_BOIMMG_347940"

            agpg140 = model.metabolites.get_by_id("1agpg140_" + c)
            agpg140.annotation["boimmg.compound"] = "C_BOIMMG_452"

            agpg141 = model.metabolites.get_by_id("1agpg141_" + c)
            agpg141.annotation["boimmg.compound"] = "C_BOIMMG_16364"

            agpg160 = model.metabolites.get_by_id("1agpg160_" + c)
            agpg160.annotation["boimmg.compound"] = "C_BOIMMG_314"

            agpg161 = model.metabolites.get_by_id("1agpg161_" + c)
            agpg161.annotation["boimmg.compound"] = "C_BOIMMG_16363"

            agpg181 = model.metabolites.get_by_id("1agpg181_" + c)
            agpg181.annotation["boimmg.compound"] = "C_BOIMMG_16317"



        ############## apg ##############

        if c == "c":

            apg140 = model.metabolites.get_by_id("apg140_" + c)
            apg140.annotation["boimmg.compound"] = "C_BOIMMG_765781"

            apg141 = model.metabolites.get_by_id("apg141_" + c)
            apg141.annotation["boimmg.compound"] = "C_BOIMMG_765700"

            apg161 = model.metabolites.get_by_id("apg161_" + c)
            apg161.annotation["boimmg.compound"] = "C_BOIMMG_765716"

            apg160 = model.metabolites.get_by_id("apg160_" + c)
            apg160.annotation["boimmg.compound"] = "C_BOIMMG_765924"

            apg181 = model.metabolites.get_by_id("apg181_" + c)
            apg181.annotation["boimmg.compound"] = "C_BOIMMG_765860"

        ############# agpe #######################

        if c == "p":
            agpe140 = model.metabolites.get_by_id("1agpe140_" + c)
            agpe140.annotation["boimmg.compound"] = "C_BOIMMG_27943"

            agpe141 = model.metabolites.get_by_id("1agpe141_" + c)
            agpe141.annotation["boimmg.compound"] = "C_BOIMMG_27925"

            agpe160 = model.metabolites.get_by_id("1agpe160_" + c)
            agpe160.annotation["boimmg.compound"] = "C_BOIMMG_291"

            agpe161 = model.metabolites.get_by_id("1agpe161_" + c)
            agpe161.annotation["seed.compound"] = "cpd25196"
            agpe161.annotation["boimmg.compound"] = "C_BOIMMG_27923"

            agpe181 = model.metabolites.get_by_id("1agpe181_" + c)
            agpe181.annotation["seed.compound"] = "cpd25197"
            agpe181.annotation["boimmg.compound"] = "C_BOIMMG_27877"

        ###################3 agpc ###############3


        # cobra.io.write_sbml_model(model,"iAF1260b_mapped.xml")
        cobra.io.write_sbml_model(model, "../case_studies/redundant_representation_case/iML1515_mapped.xml")

if __name__ == "__main__":
    map_e_coli_granulated_model()









import re

from boimmgpy.kegg.kegg_compound import KeggCompound
from boimmgpy.kegg.kegg_pathway import KeggPathway
from boimmgpy.kegg.kegg_reaction import KeggReaction
from boimmgpy.service.network_handlers.metabolic_network import MetabolicNetwork
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.model_seed.model_seed_compounds_database import ModelSeedCompoundsDB
from biocyc import biocyc
from boimmgpy.definitions import ROOT_DIR


def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█', printEnd="\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end='',flush=True)
    # Print New Line on Complete
    if iteration == total:
        print()




class PathwayHandler(object):

    def __init__(self,compounds_database = None, compounds_converter = None):
        self.target_pathways_map = {}
        self.__pathways_map = {}
        self.taxID_map = {}
        self.__home_path__ = ROOT_DIR

        if not compounds_converter:
            self.__compoundsIdConverter = CompoundsIDConverter()
        else:
            self.__compoundsIdConverter = compounds_converter

        if not compounds_database:
            self.__compounds_database = ModelSeedCompoundsDB()
        else:
            self.__compounds_database = compounds_database

        # if compounds_ontology:
        #     self.__compounds_ontology = compounds_ontology

        biocyc.set_organism('meta')

    @property
    def target_pathways_map(self):
        return self.__target_pathways_map

    @target_pathways_map.setter
    def target_pathways_map(self,target_pathways_map):
        self.__target_pathways_map = target_pathways_map

    def __set_compounds_ontology(self):
        """
        Load the compound ontology
        """
        self.__compounds_ontology = CompoundOntologyGraph()
        self.__compounds_ontology.load_ontology(self.__home_path__ + self.__configs["compounds_ontology_file"],
                                                self.__home_path__ + self.__configs["compounds_entities_file"])

    def get_targets(self, pathwayid):
        return self.target_pathways_map[pathwayid]

    def save_in_file(self,filename):
        with open(filename,"w") as file:
            for pathway in self.target_pathways_map:
                lst = self.target_pathways_map[pathway]
                new_lst = []
                temp_lst = []
                for elem in lst:
                    ms_number = elem.split("cpd")[1]
                    if int(ms_number)>100:
                        new_lst.append(elem)

                if not new_lst:
                    new_lst = temp_lst.copy()

                if new_lst:
                    file.write(">"+ ",".join(new_lst) + "\n")
                    path = self.get_path_from_pathway(pathway)

                    file.write(pathway+": " + ">".join(path) +"\n")



    def load_from_file(self,filename):
        with open(filename,"r") as file:
            line = file.readline()
            while line:
                targets = re.sub(">","",line).strip()
                targets = targets.split(",")
                line = file.readline()
                while not re.search("^>",line) and line:

                    pathwayid_path = line.split(": ")
                    path = pathwayid_path[1].strip().split(">")
                    self.target_pathways_map[pathwayid_path[0]] = targets
                    network = MetabolicNetwork()
                    if len(path) == 1:
                        network.add_vertex(path[0])
                    else:
                        for i in range(0,len(path)-1):
                            network.add_edge(path[i],path[i+1])

                    self.__pathways_map[pathwayid_path[0]] = network
                    line=file.readline()


    def get_networks_from_target(self,target_modelseedid):
        return self.target_pathways_map[target_modelseedid]

    def get_target_pathways(self):
        return self.target_pathways_map

    def get_network_from_target_and_pathway_id(self,target_modelseedid,metacyc_pathway_id):
        return self.target_pathways_map.get(target_modelseedid).get(metacyc_pathway_id)


    def add_metacyc_pathway(self,metacyc_pathway_id,generic = False):
        biocyc.set_organism('meta')
        ptw = biocyc.get(metacyc_pathway_id)

        network = self.scrap_metacyc_path(ptw,generic)
        self.convert_compounds_into_modelseedid(network)


        pathway = network.get_pathway()
        if pathway:
            target_reaction = pathway[-1]
            last_reaction = biocyc.get(target_reaction)
            products = last_reaction.compounds_right
            targets = []

            for product in products:
                modelseedids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId("MetaCyc",product.id)
                if modelseedids is not None:
                    targets.extend(modelseedids)

            self.__pathways_map[ptw.id] = network
            self.target_pathways_map[ptw.id]=targets

    def get_pathways(self):
        return self.__pathways_map

    def get_network_by_metacyc_pathway_id(self,id):
        return self.__pathways_map[id]

    def scrap_metacyc_path(self,ptw,generic):
        if generic:
            return self.__scrap_generic_metacyc_path(ptw)
        else:
            return self.__scrap_metacyc_path(ptw)

    def scrap_kegg_pathway(self,kegg_pathway):

        pathway = KeggPathway(kegg_pathway, False)
        return self.__scrap_generic_kegg_path(pathway)

    def __scrap_generic_kegg_path(self,pathway):

        network = MetabolicNetwork()

        for reaction in pathway.get_reactions():
            reaction = KeggReaction(reaction)
            reaction_id = reaction.id

            reactants = reaction.get_reactants()
            products = reaction.get_products()

            for reactant in reactants:
                reactant = KeggCompound(reactant)
                ids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId("KEGG", reactant.id)
                for id in ids:
                    compound_container = self.__compounds_database.get_compound_by_id(id)
                    if "R" in compound_container.getFormula():
                        network.add_vertex(reaction_id)
                        network.node_types["reaction"].append(reaction_id)


                        network.node_types["metabolite"].append(reactant.id)
                        network.add_vertex(reactant.id)
                        network.add_edge(reactant.id, reaction_id)
                        network.add_edge(reaction_id, reactant.id)
                        break

            for product in products:
                product = KeggCompound(product)
                ids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId("KEGG", product.id)
                for id in ids:
                    compound_container = self.__compounds_database.get_compound_by_id(id)
                    if "R" in compound_container.getFormula():
                        network.node_types["metabolite"].append(product.id)
                        network.add_vertex(product.id)
                        network.add_edge(reaction_id, product.id)
                        network.add_edge(product.id, reaction_id)
                        break

        metabolite_network = MetabolicNetwork()
        metabolite_network.convert_reaction_graph(network)
        return metabolite_network

    def __get_reactants_and_products_by_r_groups(self,reaction):

        products = reaction.get_products()
        reactants = reaction.get_reactants()

        products_r_groups = 0
        reactants_r_groups = 0

        for product in products:
            product = KeggCompound(product)
            formula = product.get_formula()
            if "R" in formula:
                match = re.match(formula, "R")
                start = match.start()
                number = formula[start+1]

                if re.match(number,"[2-9]"):
                    number= int(number)
                    products_r_groups+=number
                else:
                    products_r_groups+=1

        for reactant in reactants:
            reactant = KeggCompound(reactant)
            formula = reactant.get_formula()
            if "R" in formula:
                match = re.match(formula, "R")
                start = match.start()
                number = formula[start + 1]

                if re.match(number, "[2-9]"):
                    number = int(number)
                    reactants_r_groups += number
                else:
                    reactants_r_groups += 1

        if reactants_r_groups >= products_r_groups:

            return reactants,products

        else:
            return products,reactants

    def __scrap_metacyc_path(self,ptw):
        network = MetabolicNetwork()

        for reaction in ptw.reactions:
            reaction_id = reaction.id
            network.add_vertex(reaction_id)
            network.node_types["reaction"].append(reaction_id)

            reactants = reaction.compounds_left
            products = reaction.compounds_right

            for reactant in reactants:
                if reactant.id != "PROTON":
                    network.node_types["metabolite"].append(reactant.id)
                    network.add_vertex(reactant.id)
                    network.add_edge(reactant.id,reaction_id)

            for product in products:
                if product.id != "PROTON":
                    network.node_types["metabolite"].append(product.id)
                    network.add_vertex(product.id)
                    network.add_edge(reaction_id, product.id)

        metabolite_network = MetabolicNetwork()
        metabolite_network.convert_reaction_graph(network)
        return metabolite_network

    def __scrap_generic_metacyc_path(self,ptw):
        network = MetabolicNetwork()

        for reaction in ptw.reactions:
            reaction_id = reaction.id



            network.add_vertex(reaction_id)
            network.node_types["reaction"].append(reaction_id)

            reactants = reaction.compounds_left
            products = reaction.compounds_right

            if reaction_id == "RXN-17728":
                reactants = reaction.compounds_right
                products = reaction.compounds_left

            for reactant in reactants:
                ids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId("MetaCyc",reactant.id)
                for id in ids:
                    compound_container = self.__compounds_database.get_compound_by_id(id)
                    if "R" in compound_container.getFormula():
                        network.node_types["metabolite"].append(reactant.id)
                        network.add_vertex(reactant.id)
                        network.add_edge(reactant.id, reaction_id)
                        break


            for product in products:
                ids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId("MetaCyc", product.id)
                for id in ids:
                    compound_container = self.__compounds_database.get_compound_by_id(id)
                    if "R" in compound_container.getFormula():
                        network.node_types["metabolite"].append(product.id)
                        network.add_vertex(product.id)
                        network.add_edge(reaction_id, product.id)

        metabolite_network = MetabolicNetwork()
        metabolite_network.convert_reaction_graph(network)
        return metabolite_network

    def convert_compounds_into_modelseedid(self,network,db="MetaCyc"):
        res = {}
        # self.__babel = {}
        for compound in network.get_nodes():
            modelseedids = self.__compoundsIdConverter.convert_dbID_into_modelSeedId(db,compound)
            if modelseedids is not None:
                # self.__babel[compound] = modelseedid
                res[compound] = modelseedids
        return res


    def get_path_from_pathway(self,pathway_id):
        network = self.__pathways_map[pathway_id]
        return network.get_complete_pathway()


    def get_generic_targets_and_found_pathways(self,taxid):

        paths = self.get_pathway_by_taxID(taxid)
        targets_paths ={}
        for path in paths:
            targets = self.target_pathways_map[path]
            for target in targets:
                entity_id = self.__compounds_ontology.get_entity_id_by_model_seed_id(target)
                if entity_id:
                    # if target in targets_paths.keys():
                    targets_paths[target] = path

        targets = list(targets_paths.keys())

        res = {}

        for i in range(len(targets)):
            for j in range(i+1,len(targets)):

                entityid1 = self.__compounds_ontology.get_entity_id_by_model_seed_id(targets[i])
                entityid2 = self.__compounds_ontology.get_entity_id_by_model_seed_id(targets[j])

                container1 = self.__compounds_ontology.get_entity_container_by_id(entityid1)
                container2 = self.__compounds_ontology.get_entity_container_by_id(entityid2)

                if not container1.isGeneric() and not container2.isGeneric():
                    parent1 = self.__compounds_ontology.get_successors_by_relationship_type(entityid1,"is_a")
                    parent2 = self.__compounds_ontology.get_successors_by_relationship_type(entityid2,"is_a")

                    if parent1[0] == parent2[0]:

                        parent_modelseedid = self.__compounds_ontology.get_entity_modelseedid(parent1[0])
                        if parent_modelseedid not in res.keys():
                            res[parent_modelseedid] = [targets_paths[targets[i]]]
                            res[parent_modelseedid].append(targets_paths[targets[j]])
                        else:
                            res[parent_modelseedid].append(targets_paths[targets[i]])
                            res[parent_modelseedid].append(targets_paths[targets[j]])

        values_res = []
        for key in res:
            values_res.extend(res[key])

        values_res = self.get_unique(values_res)
        for target in targets:
            if targets_paths[target] not in values_res:
                if target not in res.keys():
                    res[target] = [targets_paths[target]]
                elif targets_paths[target] not in res[target]:
                    res[target].append(targets_paths[target])

        for key in res:
            res[key] = self.get_unique(res[key])

        return res

    def get_unique(self,lst):
        res=[]
        for elem in lst:
            if elem not in res: res.append(elem)

        return res

    def get_pathway_by_target(self,target):
        pathways = []
        for path in self.target_pathways_map:
            if target in self.target_pathways_map[path]:
                pathways.append(path)
        return pathways

    def get_pathway_by_taxID(self,taxID):
        # paths_species = []
        paths_tax_range = []
        taxa = NCBITaxa()
        lineage = taxa.get_lineage(taxID)
        for path in self.target_pathways_map:
            pathway = biocyc.get(path)

            # species = pathway.species
            # for s in species:
            #     if re.search("TAX-",s.id):
            #         id = int(s.id.replace("TAX-", ""))
            #         if id in lineage:
            #             paths_species.append((path,pathway.name))

            taxonomic_range = pathway.taxonomic_range
            for s in taxonomic_range:
                taxid = int(s.id.replace("TAX-",""))
                if taxid in lineage:
                    paths_tax_range.append(path)

        # res

        return paths_tax_range

    def get_all_paths_using_pathway_id(self,pathway_id,accept_cycles=False):
        network = self.get_network_by_metacyc_pathway_id(pathway_id)
        if not accept_cycles:
            network.prune_redundant_cycles()
        pathways = network.get_all_target_pathways()
        return pathways

    def get_path_model_seed_ids(self,ptw):
        network = self.__pathways_map[ptw]
        dict = self.convert_compounds_into_modelseedid(network)
        return dict

    def get_pathway_by_taxid_and_target(self,taxid,target):
        paths_tax_range = []
        taxa = NCBITaxa()
        lineage = taxa.get_lineage(taxid)
        for path in self.target_pathways_map:
            pathway = biocyc.get(path)

            taxonomic_range = pathway.taxonomic_range
            for s in taxonomic_range:
                taxid = int(s.id.replace("TAX-", ""))
                if taxid in lineage:
                    paths_tax_range.append(path)

        pathways = self.get_pathway_by_target(target)
        for p in paths_tax_range:
            if p in pathways:
                return p



from unittest import TestCase

from boimmgpy.database.accessors.compounds_database_accessor import set_database_information


class TestRestAccess(TestCase):

    def test_access(self):
        set_database_information(uri="bolt://palsson.di.uminho.pt:6094", user="neo4j", password="Tuning999")


from boimmgpy.database.databases_babel import BOIMMGDatabases, AliasesTransformer
from boimmgpy.database.boimmg_properties import BOIMMGProperties
from boimmgpy.database.interfaces.node import OntologyNode


class CompoundNode(OntologyNode):

    def __init__(self,node_id,node_properties,other_aliases):
        self.id = node_id
        self.extract_node_properties(node_properties,other_aliases)

    def __getattribute__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            return None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self,id):
        self._id = id


    @property
    def name(self):
        return self._name

    @name.setter
    def name(self,name):
        self._name = name


    @property
    def inchi(self):
        return self._inchi

    @inchi.setter
    def inchi(self,inchi):
        self._inchi = inchi


    @property
    def inchikey(self):
        return self._inchikey

    @inchikey.setter
    def inchikey(self,inchikey):
        self._inchikey = inchikey

    @property
    def smiles(self):
        return self._smiles

    @smiles.setter
    def smiles(self, smiles):
        self._smiles = smiles


    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, mass):
        self._mass = mass


    @property
    def generic(self):
        return self._generic

    @generic.setter
    def generic(self, generic):
        self._generic = generic


    @property
    def charge(self):
        return self._charge

    @property
    def aliases(self):
        return self._aliases

    @charge.setter
    def charge(self, value):
        self._charge = value


    @aliases.setter
    def aliases(self, aliases):
        self._aliases = aliases

    @property
    def formula(self):
        return self.__formula

    @formula.setter
    def formula(self,value):
        self.__formula = value

    @property
    def model_seed_id(self):
        return self.__model_seed_id

    @model_seed_id.setter
    def model_seed_id(self,value):
        self.__model_seed_id = value

    # def get_boimmg_aliases(self,accessor):
    #     aliases = accessor.get_all_aliases(self.id)
    #
    #     node_aliases = self.aliases



    def extract_node_properties(self, node_properties,other_aliases):

        boimmg_databases = [db.value for db in BOIMMGDatabases]
        aliases = {}
        for property in node_properties:

            if property in boimmg_databases:
                aliases[property] = [node_properties.get(property)]


            if property == BOIMMGDatabases.MODEL_SEED.value:
                self.model_seed_id = node_properties[BOIMMGDatabases.MODEL_SEED.value]

        for aliases_node in other_aliases:
            for property in aliases_node:
                if property in aliases:
                    if aliases_node[property] not in aliases[property]:
                        aliases[property].append(aliases_node[property])
                else:
                    aliases[property] = [aliases_node[property]]

        if BOIMMGProperties.FORMULA.value in node_properties:
            self.formula = node_properties[BOIMMGProperties.FORMULA.value]

        if BOIMMGProperties.MASS.value in node_properties:
            self.mass = node_properties[BOIMMGProperties.MASS.value]

        if BOIMMGProperties.SMILES.value in node_properties:
            self.smiles = node_properties[BOIMMGProperties.SMILES.value]

        if BOIMMGProperties.INCHI.value in node_properties:
            self.inchi = node_properties[BOIMMGProperties.INCHI.value]

        if BOIMMGProperties.INCHIKEY.value in node_properties:
            self.inchikey = node_properties[BOIMMGProperties.INCHIKEY.value]

        if BOIMMGProperties.CHARGE.value in node_properties:
            self.charge = node_properties[BOIMMGProperties.CHARGE.value]

        self.generic = node_properties[BOIMMGProperties.GENERIC.value]

        if BOIMMGProperties.NAME.value in node_properties:
            self.name = node_properties[BOIMMGProperties.NAME.value]

        self.aliases = AliasesTransformer.transform_boimmg_aliases_into_model_seed(aliases)






from unittest import TestCase

from boimmgpy.database.accessors.compounds_rest_accessor import CompoundsRestAccessor


class TestRestAccess(TestCase):

    def test_get_node_from_model_seed_id(self):

        accessor = CompoundsRestAccessor()

        node = accessor.get_node_from_model_seed_id("cpd27342")

        assert node.model_seed_id == "cpd27342"
        assert node.id == 98
        assert node.formula == "C8H11NO10PR2"
        assert node.inchikey == None

    def test_get_node_id_from_model_seed_id(self):

        accessor = CompoundsRestAccessor()

        id = accessor.get_node_id_from_model_seed_id("cpd27342")

        assert id == 98

    def test_get_conjugates(self):

        accessor = CompoundsRestAccessor()

        id = accessor.get_conjugates(749395)

        assert id[0] == 749396

    def test_get_predecessors_by_ont_id(self):

        accessor = CompoundsRestAccessor()

        ids = accessor.get_predecessors_by_ont_id(305731)

        assert len(ids) == 2

    def test_get_compounds_with_only_one_component(self):


        accessor = CompoundsRestAccessor()

        ids = accessor.get_compounds_with_only_one_component(11,[293815])

        assert 312156 in ids

    def test_get_compounds_with_specific_parent_within_set_of_components(self):
        accessor = CompoundsRestAccessor()

        ids = accessor.get_compounds_with_specific_parent_within_set_of_components(11, [293816,5497])
        assert 296497 in ids




import re

from boimmgpy.kegg.kegg_entity import KeggEntity


class KeggModule(KeggEntity):

    def __init__(self,entry):
        super().__init__(entry)
        self.__get_all_information()

    def get_id(self):
        return self.id

    def __get_all_information(self):
        info = self.get_raw_data().split("\n")
        self.compounds = []
        self.reactions = []
        reaction_seen = False
        compound_seen = False

        for text in info:

            if re.search("\AREACTION", text):
                reaction_seen=True
                new_line = re.sub('\s+', "\t", text )
                reaction = new_line.split("\t")[1]
                self.reactions.append(reaction)

            elif re.search("\ACOMPOUND", text):
                if reaction_seen:
                    reaction_seen = False
                    compound_seen = True
                new_line = re.sub('\s+', "\t", text)
                compound = new_line.split("\t")[1]
                self.compounds.append(compound)

            elif reaction_seen:
                new_line = re.sub('\s+', "\t", text)
                reaction = new_line.split("\t")[1]
                self.reactions.append(reaction)

            elif compound_seen:

                if re.search("\A///", text):
                    compound_seen=False

                else:
                    new_line = re.sub('\s+', "\t", text)
                    compound = new_line.split("\t")[1]
                    if re.search("\AC",compound):
                        self.compounds.append(compound)
                    else:
                        compound_seen=False

    def get_reactions(self):
        return self.reactions

    def get_compounds(self):
        return self.compounds




import Bio.KEGG.REST as kegg_api

class KeggEntity:

    def __init__(self, entry):
        self.id = entry
        self.retrieve_raw_data()

    def retrieve_raw_data(self):
        info = kegg_api.kegg_get(self.id)
        self.raw_data = ""

        for text in info.readlines():
           self.raw_data+=text

    def get_raw_data(self):
        return self.raw_data



import abc


class RepresentationProblemSolver(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'swap_and_gap_fill') and
                callable(subclass.swap_and_gap_fill) and
                hasattr(subclass, 'generalize_model') and
                callable(subclass.generalize_model) and
                hasattr(subclass, 'gap_fill_model_by_target') and
                callable(subclass.gap_fill_model_by_target) and
                hasattr(subclass, 'swap_from_generic') and
                callable(subclass.swap_from_generic)
                or
                NotImplemented)

    @abc.abstractmethod
    def swap_and_gap_fill(self, target_ontology_id):
        """This method aims at swapping all the biosynthetic precursors, and the conjugated acid and base of a given target.
        It only accepts type 1 changes."""
        raise NotImplementedError

    @abc.abstractmethod
    def generalize_model(self, target_ontology_id):
        """generalize the model"""
        raise NotImplementedError



    # @abc.abstractmethod
    # def swap_from_generic(self, target_generic_ontology_id: int, **kwargs):
    #     """
    #
    #     :param target_ontology_id:
    #     :param components:
    #     :return:
    #     """
    #     raise NotImplementedError



import re
import urllib

from rdkit import Chem

from boimmgpy.utilities import file_utilities
from boimmgpy.id_converters.id_converter import IDConverter
from boimmgpy.definitions import TOOL_CONFIG_PATH, ROOT_DIR, DATABASE_CONFIGS


class CompoundsIDConverter(IDConverter):

    def __init__(self):
        super().__init__()

        # scraper = CompoundsIDConverterScraper()
        # path = scraper()

        self.__database_config = file_utilities.read_conf_file(DATABASE_CONFIGS)
        path = ROOT_DIR + self.__database_config["path_model_seed_compounds_aliases"]
        self.__configs = file_utilities.read_conf_file(TOOL_CONFIG_PATH)
        self.__home_path__ = ROOT_DIR
        self.construct_compounds_converter(path)


    def construct_compounds_converter(self,path):

        info = open(path, "r")
        extra_info = open(self.__home_path__  + self.__configs.get("compound_converter_extra"), "r")
        url=False

        previousModelSeed = ""
        previousOtherDb = ""
        new_dic={}
        for line in info:

            if url:
                line=line.decode("utf-8")

            l = line.split("\t")


            modelSeedId = l[0]
            externalId = l[1]
            source = l[2].strip()

            if re.search("\AModelSEED", line) == None:

                if previousModelSeed != modelSeedId:
                    self.modelSeedToDb[previousModelSeed] = new_dic
                    new_dic = {}
                    new_dic[source]=[externalId]
                    previousModelSeed = modelSeedId
                    previousOtherDb = source

                elif previousOtherDb != source:
                    new_dic[source]=[externalId]
                    previousOtherDb=source

                else:
                    new_dic[source].append(externalId)

                if source in self.dbToModelSeed.keys():
                    if externalId in self.dbToModelSeed[source].keys():
                        self.dbToModelSeed[source][externalId].append(modelSeedId)
                    else:
                        self.dbToModelSeed[source][externalId] = [modelSeedId]

                else:
                    dic = {}
                    dic[externalId] = [modelSeedId]
                    self.dbToModelSeed[source] = dic
        info.close()

        for line in extra_info:

            if url:
                line=line.decode("utf-8")

            l = line.split("\t")


            modelSeedId = l[0]
            externalId = l[1]
            source = l[2].strip()

            if re.search("\AModelSEED", line) == None:

                if modelSeedId in self.modelSeedToDb.keys():
                    if source in self.modelSeedToDb.get(modelSeedId) and \
                            externalId not in self.modelSeedToDb[modelSeedId][source]:
                        self.modelSeedToDb[modelSeedId][source].append(externalId)
                    else:
                        self.modelSeedToDb[modelSeedId][source] = [externalId]

                if source in self.dbToModelSeed.keys():
                    if externalId in self.dbToModelSeed[source].keys():
                        self.dbToModelSeed[source][externalId].append(modelSeedId)
                    else:
                        self.dbToModelSeed[source][externalId] = [modelSeedId]

                else:
                    dic = {}
                    dic[externalId] = [modelSeedId]
                    self.dbToModelSeed[source] = dic

        extra_info.close()


    def convert_molfile_from_kegg_into_inchikey(self,compound_id):
        try:
            data = urllib.request.urlopen("https://www.genome.jp/dbget-bin/www_bget?-f+m+compound+"+compound_id)
        except:
            print("error connecting")
        molformat = ""

        for line in data:
            molformat+=line.decode('utf-8')

        try:
            mol = Chem.MolFromMolBlock(molformat)
            inchikey = Chem.MolToInchiKey(mol)
        except:
            return None
        return inchikey

    def find_in_model_seed_by_inchikey(self,inchikey,modelseed_database):
        compound_id = ""
        for key in modelseed_database:
            if key==inchikey:
                compound_id=modelseed_database[key].get("model_seed_id")
                break
        return compound_id


import requests
from boimmgpy import definitions
from boimmgpy.utilities import file_utilities


class RestUtils:

    @staticmethod
    def map_model(model,database):

        conf = file_utilities.read_conf_file(definitions.BOIMMG_DATABASE)
        rest_uri = conf["rest_uri"]

        with open(model, 'rb') as file:

            file = {'model': file}
            response = requests.post(rest_uri + "rest/map_model/" + database, files = file)

        return response


import abc
import time

import requests
from neo4j import GraphDatabase

from boimmgpy import definitions
from boimmgpy.model_seed.model_seed_compound import ModelSeedCompound
from boimmgpy.utilities import file_utilities


class ModelSeedCompoundsDB:

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'login') and
                callable(subclass.login) and
                hasattr(subclass, 'get_predecessors_by_ont_id') and
                callable(subclass.get_predecessors_by_ont_id) and
                hasattr(subclass, 'get_predecessors_by_ont_id_rel_type') and
                callable(subclass.get_predecessors_by_ont_id_rel_type) and
                hasattr(subclass, 'get_node_by_ont_id') and
                callable(subclass.get_node_by_ont_id)
                or
                NotImplemented)

    @abc.abstractmethod
    def get_compound_by_id(self, ont_id: int) -> list:
        """Get predecessors using as parameter the database identifier"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_compound_by_inchi_key(self, inchikey: str) -> list:
        """Get predecessors using as parameter the database identifier"""
        raise NotImplementedError

class ModelSeedCompoundsDBRest(ModelSeedCompoundsDB):

    def __init__(self):

        conf = file_utilities.read_conf_file(definitions.BOIMMG_DATABASE)
        self.rest_uri = conf["rest_uri_model_seed"]

        self.sleep_time = 0.5

    def deserialize_compound(self,compound_dict):

        return ModelSeedCompound(compound_dict)

    def get_compound_by_id(self, id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "db_id/" + id)

        return self.deserialize_compound(res)


    def get_compound_by_inchi_key(self, inchikey):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "inchi_key/" + inchikey)

        return self.deserialize_compound(res)


class ModelSeedCompoundsDBRaw(ModelSeedCompoundsDB):

    def __init__(self,uri="",user="",password=""):


        if not uri or not user or not password:
            uri, user, password = self.read_config_file()

        self.__uri = uri
        self.__user = user
        self.__password = password

    def read_config_file(self):
        configs = file_utilities.read_conf_file(definitions.BOIMMG_DATABASE)

        uri = configs["uri"]
        user = configs["user"]
        password = configs["password"]

        return uri, user, password

    def login(self):

        driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__password), encrypted=False)
        self.tx = driver

    @property
    def tx(self):
        return self._tx

    @tx.setter
    def tx(self,tx):
        self._tx = tx



    def get_compound_by_id(self,id):

        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:ModelSeedCompound) "
                                 "WHERE c.model_seed_id = $db_id "
                                 "RETURN p",
                                 db_id=id)

            node = result.single()

            if node:
                node_properties = node.get("p")

                return ModelSeedCompound(node_properties)

        return None


    def get_compound_by_inchi_key(self,inchikey):

        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:ModelSeedCompound) "
                                 "WHERE c.inchikey contains $inchikey "
                                 "RETURN c "
                                 ,
                                 inchikey=inchikey[:-1])
            data = result.single()
            if data:
                node_properties = data.get("p")

                return ModelSeedCompound(node_properties)


        return None

    # def get_compound_by_canonical_smiles(self,smiles):
    #     if smiles in self.__smiles_database.keys():
    #         return self.__smiles_database[smiles]
    #     return None

    # def get_compounds(self):
    #     return self.__compounds

    # def find_compound_by_inchikey(self,inchikey):
    #     for compound in self.__compounds_database:
    #         compound_container = self.__compounds_database[compound]
    #         db_compound_inchikey = compound_container.getInchikey()
    #         if db_compound_inchikey:
    #             if db_compound_inchikey[:-2] == inchikey[:-2]:
    #                 return compound_container
    #     return None


    # def __set_inchi_key_and_smiles_database(self,compound):
    #
    #     inchikey = compound.getInchikey()
    #     smiles = compound.getSmiles()
    #
    #     if inchikey or smiles:
    #
    #         self.__inchikey_database[inchikey[:-1]] = compound
    #         mol = MolFromSmiles(smiles)
    #         if mol:
    #
    #             canonical_smiles_not_isomer = MolToSmiles(mol,isomericSmiles=True)
    #
    #             canonical_smiles = MolToSmiles(mol)
    #
    #             try:
    #                 neutralized, _ = chemo_utilities.NeutraliseCharges(canonical_smiles)
    #
    #                 neutralized_not_isomer, _ = chemo_utilities.NeutraliseCharges(canonical_smiles_not_isomer)
    #
    #                 if neutralized in self.__smiles_database:
    #
    #                     self.__smiles_database[neutralized].append(compound)
    #
    #                 else:
    #
    #                     self.__smiles_database[neutralized] = [compound]
    #
    #                 if neutralized_not_isomer in self.isomer_smiles_database:
    #
    #                     self.isomer_smiles_database[neutralized_not_isomer].append(compound)
    #
    #                 else:
    #
    #                     self.isomer_smiles_database[neutralized_not_isomer] = [compound]
    #
    #
    #             except:
    #                 pass
    #
    #
    #
    # def __read_model_seed_structures(self, path):
    #     with open(path,"r") as f:
    #         f.readline()
    #         compounds = f.readlines()
    #         previous = compounds[0].split("\t")[0]
    #         mapped_features = {}
    #         for compound in compounds:
    #             features = compound.split("\t")
    #             id = features[0]
    #             if previous != id:
    #                 smiles = mapped_features["smile"]
    #                 if "inchikey" in mapped_features.keys():
    #                     inchikey = mapped_features["inchikey"]
    #
    #                 if "inchi" in mapped_features.keys():
    #                     inchi = mapped_features["inchi"]
    #
    #                 self.__compounds_database[previous].setStructures(smiles,inchikey,inchi)
    #
    #                 self.__set_inchi_key_and_smiles_database(self.__compounds_database[previous])
    #
    #                 mapped_features = {}
    #             mapped_features[features[1].lower()] = features[5].strip()
    #             previous = id
    #
    #
    #
    # def __read_model_seed_compounds_database(self, path):
    #     self.__compounds_database = {}
    #     with open(path,"r") as f:
    #         f.readline()
    #         compounds = f.readlines()
    #         for compound in compounds:
    #
    #             features = compound.split("\t")
    #
    #             name = features[2]
    #             formula = features[3]
    #             charge = int(features[7])
    #             inchikey = features[6]
    #             smiles = features[19]
    #             modelseedCompound = ModelSeedCompound(features[0],name,charge,formula)
    #             modelseedCompound.setStructures(smiles,inchikey)
    #
    #
    #
    #             self.__compounds_database[features[0]] = modelseedCompound
    #             self.__compounds.append(modelseedCompound)





from enum import Enum


class BOIMMGDatabases(Enum):
    HMDB = "hmdb_id"
    LIPID_MAPS = "lipid_maps_id"
    MODEL_SEED = "model_seed_id"
    METACYC = "metacyc_id"
    SWISS_LIPIDS = "swiss_lipids_id"
    KEGG_ID = "kegg_id"
    CHEBI = "chebi_id"
    METANETX_ID = "metanetx_id"
    BIGG = "bigg_id"
    PUBMED = "pubmed_cid"

class ModelSEEDDatabases(Enum):
    MODEL_SEED = "ModelSEED"
    BIGG = "BiGG"
    KEGG_ID = "KEGG"
    METACYC = "MetaCyc"
    METANETX_ID = "metanetx.chemical"
    SWISS_LIPIDS = "SWISS_LIPIDS"
    LIPID_MAPS = "LIPID_MAPS"

class AliasesTransformer:

    @staticmethod
    def transform_dictionary_in_boimmg_databases(aliases):
        boimmg_format_aliases = {}
        for database in ModelSEEDDatabases:
            if database.value in aliases:
                alias_lst = aliases.get(database.value)
                boimmg_alias = BOIMMGDatabases[database.name].value

                boimmg_format_aliases[boimmg_alias] = alias_lst[0]

        return boimmg_format_aliases

    @staticmethod
    def transform_boimmg_aliases_into_model_seed(aliases):
        boimmg_format_aliases = {}
        model_seed_dbs = [database.name for database in ModelSEEDDatabases]

        for database in BOIMMGDatabases:
            if database.value in aliases:
                alias_lst = aliases.get(database.value)
                if database.name in model_seed_dbs:
                    boimmg_alias = ModelSEEDDatabases[database.name].value

                    boimmg_format_aliases[boimmg_alias] = alias_lst

        return boimmg_format_aliases

    @staticmethod
    def convert_model_seed_format_into_boimmg(database_name: str) -> str:
        for database in ModelSEEDDatabases:
            if database.value == database_name:
                return BOIMMGDatabases[database.name].value

        return



from cobra import Model, Reaction

from boimmgpy.service.revisor.compounds_revisor import CompoundsRevisor
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import TOOL_CONFIG_PATH,ROOT_DIR,REACTIONS_ANNOTATION_CONFIGS_PATH,COMPOUNDS_ANNOTATION_CONFIGS_PATH

class ReactionsChanger:

    def __init__(self, model : Model, type:int , model_database : str,
                 universal_model: Model,compounds_converter=None):
        """
        Class constructor

        :param cobrapy.Model model: COBRApy model
        :param int type: type of change (0,1,2 or 3)
        :param string model_database: database format of metabolites and reactions
        :param cobrapy.Model universal_model: universal model
        :param (optional) ReactionsIDConverter reactions_converter: a reaction identifier converter
        """

        if not compounds_converter:
            self.__compoundsIDConverter = CompoundsIDConverter()

        else:
            self.__compoundsIDConverter = compounds_converter

        if not universal_model:
            self.__universal_model = Model()
        else:
            self.__universal_model = universal_model

        self.report_material = {}
        self.__type = type
        self.__model_database = model_database
        self.__model = model
        self.__modelseed_hydrogen = "cpd00067"
        self.__reactionBalancer = CompoundsRevisor(self.__model)
        self.__not_found_reactions_num = 0

        self.__home_path__ = ROOT_DIR

        self.__configs = file_utilities.read_conf_file(
            TOOL_CONFIG_PATH)
        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(
            COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)


    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self,value):
        self.__model = value


    def swap_reactions_by_compounds(self, compounds_with_reactions_to_change : list) -> list:
        """
        This method swap all the reactions associated to a list of compounds.
        Firstly, it tries to balance the reactions and then tries to change the reaction.

        :param list compounds_with_reactions_to_change: list of Model.metabolites with reactions to change.
        :return list: reactions changed
        """
        past = []
        reaction_changed = []
        for compound in compounds_with_reactions_to_change:
            reactions = self.get_reactions_by_compound(compound)
            for reaction in reactions:
                if reaction not in past and reaction.annotation.get("sbo") != "SBO:0000629" \
                        and "Biomass" not in reaction.id:

                    past.append(reaction)

                    self.change_reaction(reaction)

                    reaction_changed.append(reaction)

        return reaction_changed

    def get_reactions_by_compound(self,compound):
        """
        Return all the reactions associated with a given compound

        :param Metabolite compound:
        :return list<Reaction>: all the reactions associated with a given compound
        """
        res=[]
        for reaction in self.__model.reactions:
            metabolites_ids = [m.id for m in reaction.metabolites]

            if compound.id in metabolites_ids:
                res.append(reaction)

        return res

    def swap_reactions(self, reactions_to_change) -> list:
        """
        This method swap all the reactions associated to a list of compounds.
        Firstly, it tries to balance the reactions and then tries to change the reaction.

        :param list reactions_to_change: list of Model.reactions with reactions to change.
        :return list: list of changed reactions
        """
        past = []
        reaction_changed = []
        for reaction in reactions_to_change:
            if reaction not in past and reaction.annotation.get("sbo") != "SBO:0000629" and "Biomass" not in reaction.id:
                past.append(reaction.copy())

                self.change_reaction(reaction)

                reaction_changed.append(reaction)

        return reaction_changed


    def change_reaction(self, reaction):
        """
        This method changes a given reaction's name, id, delta G, etc...
        It is worth noting that the reaction in the model is no longer the true one, since the metabolites were swapped
        1º start by searching in the ontology whether there is any entity representing the reaction in the model
        2º - if the reaction is found then other reaction in the ontology is searched to replace it
        3º - otherwise the reaction will be searched in a reaction database
        4º - if the reaction was not found, then a cannonical name and id will be set.

        :param Model.reaction reaction: reaction to be changed.

        """

        # if self.__model_database in self.__reactionsAnnotationConfigs:
        #     database = self.__reactionsAnnotationConfigs[self.__model_database]
        # else:
        #     raise Exception


        self.__change_not_found_reaction(reaction)

    def __change_not_found_reaction(self, reaction):
        """
        This method will change a not found reaction in the database

        :param Model.reaction reaction: reaction to be changed.

        """

        self.change_boimmg_reaction_id(reaction)

        reaction.name = "Changed - "+ reaction.name
        reaction.annotation = {}

    def change_boimmg_reaction_id(self,reaction : Reaction):

        new_id = self.__reactionsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]
        metabolites = reaction.metabolites

        metabolites = sorted([metabolite.id for metabolite in metabolites])
        new_id += "_".join(metabolites)

        self.report_material[reaction.id] = new_id
        reaction.id = new_id

        return reaction.id


    def get_database_ids_compounds_in_model(self,compounds_in_model):
        """
        This method searchs in a given database the ids of the compounds in the model

        :param dict compounds_in_model: model compounds assigned to a model reaction

        :returns list: list with database ids of the metabolites assigned to a model reaction

        """

        res = []
        for compound in compounds_in_model.keys():
            modelseed_id_annotation = compound.annotation.get("seed.compound")

            if modelseed_id_annotation and type(modelseed_id_annotation) == str:
                modelseed_id=[modelseed_id_annotation]

            else:
                modelseed_id=modelseed_id_annotation

            if self.__model_database!="ModelSEED" and not modelseed_id:
                db_id = compound.annotation.get(self.__compoundsAnnotationConfigs[self.__model_database])
                if type(db_id) == list:

                    i=0
                    found=False

                    while not found and i<len(db_id):
                        modelseed_id = self.__compoundsIDConverter.convert_dbID_into_modelSeedId(self.__model_database,
                                                                                                 db_id[i])

                        if modelseed_id:
                            found=True
                else:
                    modelseed_id = self.__compoundsIDConverter.convert_dbID_into_modelSeedId(self.__model_database,
                                                                                             db_id)

            if not modelseed_id:
                return []

            else:
                if not res:
                    for id in modelseed_id:
                        res.append([id])
                else:
                    if len(modelseed_id) == 1:
                        for reaction in res:
                            reaction.append(modelseed_id[0])
                    else:
                        previous = res.copy()
                        res=[]
                        for id in modelseed_id:
                            for reaction in previous:
                                new_reaction = reaction.copy()
                                new_reaction.append(id)
                                res.append(new_reaction)

        final_res = []
        for reaction in res:
            final_res.append(sorted(reaction))
        return final_res

    def set_type2_is_generic_in_model(self, isGenericInModel):
        """
        This method sets the state of the metabolite in the model.

        :param boolean isGenericInModel: if True -> Generic; if False -> Complete
        :return:
        """
        self.__set_type2_is_generic_in_model = isGenericInModel

    def set_type(self, new_type):

        """
        This method sets the type of swap.
        :param int new_type: Type of swap (0,1,2,3)
        :return:
        """

        self.__type = new_type







import os

import setuptools
from boimmgpy import definitions

with open("boimmgpy/README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="boimmg",
    version="0.1.10",
    author="João Capela",
    author_email="jcapels96@gmail.com",
    description="Find and tag Git commits based on version numbers in commit messages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://initialcommit.com/projects/git-tagup",
    packages=setuptools.find_packages(),
    include_package_data=True,
    # package_dir={'project_initializer': 'boimmgpy'},
    # package_data={'project_initializer':
    #     os.listdir('configs/') +
    #      os.listdir("required_files/biochemistry/Aliases/")}

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'gitpython',
        'pandas',
        'cobra',
        'cobrababel',
        'neo4j',
        'sympy',
        'biocyc',
        'biopython',
        'requests',
        'numpy',
        'python-libsbml'
    ],

    keywords='git tag boimmg tagup tag-up version autotag auto-tag commit message',
    project_urls={
        'Homepage': 'https://github.com/jcapels/boimmg',
    },
    # entry_points={
    #     'console_scripts': [
    #         'git-tagup=git_tagup.__main__:main',
    #         'gtu=git_tagup.__main__:main',
    #     ],
    # },
)


import math

import cobra
import re
from sympy.solvers import solve
from collections import defaultdict

from boimmgpy import definitions


def map_iJR904():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iJR904.xml")

    cardiolipin = model.metabolites.get_by_id("clpn_EC_c")

    phosphatidylserine = model.metabolites.get_by_id("ps_EC_c")

    phosphatidylethanolamine = model.metabolites.get_by_id("pe_EC_c")

    phosphatidylglycerol =  model.metabolites.get_by_id("pg_EC_c")

    phosphatidylglycerophosphate =  model.metabolites.get_by_id("pgp_EC_c")

    CDPdiacylglycerol = model.metabolites.get_by_id("cdpdag_EC_c")

    phosphatidate = model.metabolites.get_by_id("pa_EC_c")

    diacylglycerol = model.metabolites.get_by_id("12dgr_EC_c")

    acyl_glycerophosphoglycerol = model.metabolites.get_by_id("agpg_EC_c")

    acyl_phosphatidylglycerol = model.metabolites.get_by_id("apg_EC_c")

    acyl_glycerophosphoethanolamine = model.metabolites.get_by_id("agpe_EC_c")

    acyl_glycerophosphocholine = model.metabolites.get_by_id("agpc_EC_c")






    ################## mapping ###################

    acyl_glycerophosphocholine.annotation = {}
    acyl_glycerophosphocholine.annotation["seed.compound"] = "cpd21879"

    acyl_phosphatidylglycerol.annotation = {}
    acyl_phosphatidylglycerol.annotation["seed.compound"] = "cpd15649"

    acyl_glycerophosphoglycerol.annotation = {}
    acyl_glycerophosphoglycerol.annotation["seed.compound"] = "cpd19396"

    acyl_glycerophosphoethanolamine.annotation = {}
    acyl_glycerophosphoethanolamine.annotation["seed.compound"] = "cpd12547"

    cardiolipin.annotation = {}
    cardiolipin.annotation["seed.compound"] = "cpd22513"

    phosphatidylserine.annotation = {}
    phosphatidylserine.annotation["seed.compound"] = "cpd27342"

    phosphatidylethanolamine.annotation = {}
    phosphatidylethanolamine.annotation["seed.compound"] = "cpd27339"

    phosphatidylglycerol.annotation = {}
    phosphatidylglycerol.annotation["seed.compound"] = "cpd27340"

    phosphatidylglycerophosphate.annotation = {}
    phosphatidylglycerophosphate.annotation["seed.compound"] = "cpd27341"

    CDPdiacylglycerol.annotation = {}
    CDPdiacylglycerol.annotation["seed.compound"] = "cpd22517"

    phosphatidate.annotation = {}
    phosphatidate.annotation["seed.compound"] = "cpd27368"

    diacylglycerol.annotation = {}
    diacylglycerol.annotation["seed.compound"] = "cpd26855"

    # return model
    cobra.io.write_sbml_model(model, "iJR904_mapped.xml")

def balance_reaction(eq):
    Ls=list('abcdefghijklmnopqrstuvwxyz')

    Ss,Os,Es,a,i=defaultdict(list),Ls[:],[],1,1

    for p in eq.split('->'):
        for k in p.split('+'):
            c = [Ls.pop(0), 1]
            for e,m in re.findall('([A-Z][a-z]?)([0-9]*)',k):
                m=1 if m=='' else int(m)
                a*=m
                d=[c[0],c[1]*m*i]
                Ss[e][:0],Es[:0]=[d],[[e,d]]
        i=-1

    Ys=dict((s,eval('Symbol("'+s+'")')) for s in Os if s not in Ls)

    Qs=[eval('+'.join('%d*%s'%(c[1],c[0]) for c in Ss[s]),{},Ys) for s in Ss]+[Ys['a']-a]

    k=solve(Qs,*Ys)

    if k:
        N=[k[Ys[s]] for s in sorted(Ys)]

        g=N[0]

        for a1, a2 in zip(N[0::2],N[1::2]):
            g=math.gcd(g,a2)

        N=[i/g for i in N]

        pM=lambda c: str(c) if c!=1 else ''

        print('->'.join('+'.join(pM(N.pop(0))+str(t) for t in p.split('+')) for p in eq.split('->')))
    else: print('Nope!')


if __name__=="__main__":
    # balance_reaction("H2O+C38H73O13P2+H->C38H75O10P+HO4P")
    map_iJR904()








import json
import re
import time
from copy import deepcopy

import cobra
from biocyc import biocyc
from cobra import Model, Reaction
from cobra.util import linear_reaction_coefficients

from boimmgpy import definitions
from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.service.network_modifiers.granulator import Granulator
from boimmgpy.service.revisor.compounds_revisor import CompoundsRevisor
from boimmgpy.service.model_mapper import ModelMapper
from boimmgpy.service.interfaces.representation_problem_solver import RepresentationProblemSolver
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.model_seed.model_seed_compounds_database import ModelSeedCompoundsDBRaw, \
    ModelSeedCompoundsDBRest
from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import ROOT_DIR, TOOL_CONFIG_PATH, COMPOUNDS_ANNOTATION_CONFIGS_PATH, REACTIONS_ANNOTATION_CONFIGS_PATH
from boimmgpy.utilities.rest_access_utils import RestUtils

PROGRESS_BAR = ROOT_DIR + "/service/logs/progress_bar.txt"
# PROGRESS_BAR = "/workdir/resultsWorker/progress_bar.txt"

class RedundantCaseSolver(RepresentationProblemSolver):

    def __init__(self, model, database_format,db_accessor = CompoundsDBAccessor()):
        """
        Class constructor

        :param Model model: cobrapy model
        :param str database_format: ModelSEED, BiGG or KEGG
        """

        self.model = model

        self.objective = linear_reaction_coefficients(self.model)
        if self.objective:
            self.objective = list(self.objective.keys())[0]
        else:
            raise Exception("No objective found")

        self.__compounds_ontology = db_accessor

        self.model = model
        self.__database_format = database_format
        self.__configs = file_utilities.read_conf_file(TOOL_CONFIG_PATH)
        self.__home_path__ = ROOT_DIR
        self.__define_instance_variables()

        if isinstance(db_accessor, CompoundsDBAccessor):
            self.__modelseedCompoundsDb = ModelSeedCompoundsDBRaw()

        else:
            self.__modelseedCompoundsDb = ModelSeedCompoundsDBRest()


        self.write_in_progress_bar("mapping model... ",1)

        self.write_in_progress_bar("model mapped ", 10)


    def map_model(self):
        """
        Creates maps for all the metabolites in the model

        :return:
        """

        self.write_in_progress_bar("mapping the model... ", 10)

        if isinstance(self.__compounds_ontology, CompoundsDBAccessor):
            self.__mapper.map_model(self.__database_format)

        else:
            cobra.io.write_sbml_model(self.__model, definitions.ROOT_DIR + "/temp/model_to_be_submitted.xml")
            response = RestUtils.map_model(definitions.ROOT_DIR + "/temp/model_to_be_submitted.xml", self.__database_format)
            maps = response.json().get("result")

            self.__mapper.boimmg_db_model_map = maps["boimmg_db_model_map"]
            self.__mapper.boimmg_db_model_map_reverse = maps["boimmg_db_model_map_reverse"]
            self.__mapper.compounds_aliases_indexation = maps["compounds_aliases_indexation"]
            self.__mapper.compounds_aliases_indexation_reverse = maps["compounds_aliases_indexation_reverse"]
            self.__mapper.compound_inchikey_indexation = maps["compound_inchikey_indexation"]

        self.__mapper.mapped = True


        self.write_in_progress_bar("model mapped ", 31)

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, value):
        self.__model = value

    @property
    def mapper(self):
        return self.__mapper

    @mapper.setter
    def mapper(self,value):
        self.__mapper = value

    def write_in_progress_bar(self, message, value):
        with open(PROGRESS_BAR,"w") as file:
            file.write(message + ":" + str(value))

    def dump_maps(self,destination:str):
        """
        It creates a dump of all the model metabolite maps

        :param str destination: destination folder
        :return:
        """

        self.__mapper.create_map_dump(destination)

    def load_maps(self,folder):
        """
        Load all the metabolite maps in a given folder

        :param str folder: folder path
        :return:
        """

        self.__mapper.upload_maps(folder)


    def __define_instance_variables(self):
        """
        Define instance variables

        :return:
        """

        self.__universal_model = Model("universal_model")
        self.__swapped = []

        #self.__compounds_ontology = CompoundsDBAccessor()

        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(
            COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)


        self.__compoundsIdConverter = CompoundsIDConverter()

        self.__compounds_revisor = CompoundsRevisor(self.model, self.__universal_model)

        biocyc.set_organism("meta")

        self.mapper = ModelMapper(self.model,self.__compoundsIdConverter,self.__compounds_ontology)

        self.__compounds_revisor.set_model_mapper(self.mapper)

        self.granulator = Granulator(self.model, self.mapper, self.__database_format, self.__compoundsIdConverter,
                                     self.__compoundsAnnotationConfigs)

    def __add_hydrogen_to_virtual_model(self):
        """
        This method set searches for the hydrogen in each compartment of the model.
        The hydrogens are used to balance reactions.
        """


        compartments = self.__model.compartments
        metabolites_in_model = self.__model.metabolites

        res = []

        i = 0
        while i<len(metabolites_in_model):

            metabolite = metabolites_in_model[i]

            if metabolite.formula != "" and metabolite.formula in ["H","H+"] and metabolite.compartment in compartments:
                res.append(metabolite)

            i+=1

        if res:
            self.__virtual_model.add_metabolites(res)

        else:
            raise Exception("Please insert formulas to the hydrogens or protons in your model")

    def swap_and_gap_fill(self,target_ontology_id):
        pass

    def generalize_model(self,target_ontology_id):
        pass

    def gap_fill_model_by_target(self,target_ontology_id: int, components_ont_ids: list):
        pass

    def write_reports(self,file_path):

        metabolites_report_material = self.granulator.metabolite_report_material
        reactions_report_material = self.granulator.reaction_report_material

        with open(file_path, "w") as f:

            f.write("--Metabolites--\n")
            f.write("metabolite in model,replacer metabolite\n")

            for metabolite in metabolites_report_material:
                f.write(metabolite + ",")
                f.write(str(metabolites_report_material[metabolite]) + "\n")

            f.write("--Reactions--\n")
            f.write("reaction in model,replacer reaction\n")

            for reaction in reactions_report_material:
                f.write(reaction + ",")
                f.write(str(reactions_report_material[reaction]) + "\n")

    def swap_from_generic(self,targets: list, components: list,same_components = False,
                          progress_bar_processes_left = 1):
        """
        Swap compounds from a generic target and components. The algorithm acts as follows:

            1st: Starts for identifying the reactions present in the requested biosynthetic pathway;

            2nd: Creates a virtual model;

            3rd: Calls the granulator to granulate the biosynthetic pathway;

        :param list<str> targets: list of lipid targets
        :param list<str> components: list of components
        :param bool same_components: boolean for same or mix of components
        :param int progress_bar_processes_left:
        :return:
        """

        if not self.mapper.mapped:
            self.map_model()

        targets_generic_ontology_id, components_ont_ids = \
            self.transform_targets_and_components_into_boimmg_ids(targets,components)

        self.__virtual_model = Model("virtual_model")

        self.__add_hydrogen_to_virtual_model()

        self.__virtual_model_mapper = ModelMapper(self.__virtual_model,
                                                self.__compoundsIdConverter,self.__compounds_ontology)

        self.__virtual_compounds_revisor = CompoundsRevisor(self.__virtual_model, self.__universal_model,
                                                            self.__compoundsAnnotationConfigs)

        self.__virtual_compounds_revisor.set_model_mapper(self.__virtual_model_mapper)



        self.granulator.set_virtual_model(self.__virtual_model,self.__virtual_model_mapper,self.__virtual_compounds_revisor)


        for target_generic_ontology_id in targets_generic_ontology_id:


            biosynthesis_reactions = self.granulator.identify_biosynthesis_pathway(target_generic_ontology_id)


            self.__biosynthesis_reactions = deepcopy(biosynthesis_reactions)

            self.__add_new_reactions_to_virtual_model(self.__biosynthesis_reactions)

            self.__model.remove_reactions(biosynthesis_reactions)

            self.write_in_progress_bar("starting granulation ", 32)
            self.granulator.granulate(target_generic_ontology_id, components_ont_ids, same_components,
                                progress_bar_processes_left)

        self.generateISAreactions()

    def __subtract_reactions_in_list(self,reactions,reactions_to_subtract):

        res = []

        for reaction in reactions:
            if reaction.id not in reactions_to_subtract:
                res.append(reaction)

        return res

    # def solve_components_reactions(self,same_components=False):
    #
    #     self.__virtual_model = Model("virtual_model")
    #     self.__universal_model = Model("universal_model")
    #
    #     self.__virtual_model_mapper = ModelMapper(self.__virtual_model, self.__compoundsAnnotationConfigs,
    #                                               self.__compoundsIdConverter)
    #
    #     self.__virtual_compounds_revisor = CompoundsRevisor(self.__virtual_model, self.__universal_model,
    #                                                         self.__compoundsAnnotationConfigs)
    #
    #     self.__virtual_compounds_revisor.set_model_mapper(self.__virtual_model_mapper)
    #
    #     granulator = Granulator(self.model, self.mapper, self.__database_format, self.__modelseedCompoundsDb,
    #                             self.__compoundsIdConverter, self.__compoundsAnnotationConfigs)
    #
    #     granulator.set_virtual_model(self.__virtual_model, self.__virtual_model_mapper,
    #                                  self.__virtual_compounds_revisor)
    #
    #     granulator.solve_components_reactions(same_components)


    def __add_new_reactions_to_virtual_model(self,new_reactions):
        self.__virtual_model_mapper.add_new_reactions_to_model(new_reactions)
        self.__virtual_model.add_reactions(new_reactions)

    def __add_new_reactions_to_model(self,new_reactions):
        self.mapper.add_new_reactions_to_model(new_reactions)
        self.model.add_reactions(new_reactions)

    def generateISAreactions(self):

        parents_and_children = self.get_parent_and_children_in_model()

        objective_reactants = [reactant.id for reactant in self.objective.reactants]

        reactions = []
        for parent in parents_and_children:

            if parent in objective_reactants:

                for child in parents_and_children[parent]:

                    name = "ISA_reaction_"+child

                    id = "ISA_reaction_"+child

                    newISAreaction = Reaction(id = id, name = name)

                    child_cobra_container = self.model.metabolites.get_by_id(child)

                    parent_cobra_container = self.model.metabolites.get_by_id(parent)

                    stoichiometry = {child_cobra_container: -1,
                                     parent_cobra_container: 1}

                    newISAreaction.add_metabolites(stoichiometry)

                    reactions.append(newISAreaction)

        self.model.add_reactions(reactions)


    def get_parent_and_children_in_model(self):

        res = {}

        for model_compound_id in self.mapper.boimmg_db_model_map.keys():

            boimmg_id = self.mapper.boimmg_db_model_map[model_compound_id]
            node = self.__compounds_ontology.get_node_by_ont_id(boimmg_id)

            if not node.generic:
                parents = self.__compounds_ontology.get_successors_by_ont_id_rel_type(boimmg_id,"is_a")

                if parents:

                    parent = parents[0]
                    if parent in self.mapper.boimmg_db_model_map_reverse.keys():
                        parents_in_model = self.mapper.boimmg_db_model_map_reverse[parent]

                        if len(parents_in_model) == 1:
                            parent_in_model = parents_in_model[0]

                            if parent_in_model in res:
                                res[parent_in_model].append(model_compound_id)

                            else:
                                res[parent_in_model] = [model_compound_id]
        return res


    def transform_targets_and_components_into_boimmg_ids(self, targets, components):

        """
        Conversion of components and lipid targets into BOIMMG's format

        :param targets:
        :param components:
        :return:
        """

        targets_res = []
        components_res = []
        accessor = CompoundsDBAccessor()
        boimmg_prefix = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]

        for target in targets:
            new_target = self._transform_database_ids_into_boimmg_ids(target,accessor,boimmg_prefix)

            if not new_target:
                exit(2)

            targets_res.append(new_target)

        for component in components:
            new_component = self._transform_database_ids_into_boimmg_ids(component, accessor, boimmg_prefix)

            if not new_component:
                exit(2)

            components_res.append(new_component)

        return targets_res,components_res


    def _transform_database_ids_into_boimmg_ids(self, target: str, accessor: CompoundsDBAccessor, boimmg_prefix: str):

        """
        Convert target lipid identifier into BOIMMG's format

        :param str target:
        :param CompoundsDBAccessor accessor:
        :param str boimmg_prefix:
        :return:
        """

        targets_res = ''
        if boimmg_prefix in target:
            targets_res = int(target.replace(boimmg_prefix, ""))


        elif re.match("^[0-9]*$", target):

            targets_res = int(target)

        elif "cpd" not in target:
            model_seed_ids = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(target)

        else:
            model_seed_ids = [target]

        if not targets_res:
            for model_seed_id in model_seed_ids:
                node1 = accessor.get_node_from_model_seed_id(model_seed_id)

                if node1:
                    targets_res = node1.id
                    break

        return targets_res

































import abc

class OntologyNode(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'extract_node_properties') and
                callable(subclass.extract_node_properties) and
                hasattr(subclass, 'id') and
                hasattr(subclass, 'name') and
                hasattr(subclass, 'inchi') and
                hasattr(subclass, 'inchikey') and
                hasattr(subclass, 'smiles') and
                hasattr(subclass, 'mass') and
                hasattr(subclass, 'generic') and
                hasattr(subclass, 'charge') and
                hasattr(subclass, 'aliases') and
                hasattr(subclass, 'formula')
                or
                NotImplemented)

    @abc.abstractmethod
    def extract_node_properties(self,node_properties: dict,other_aliases: dict):
        raise NotImplementedError


import re

from boimmgpy.model_seed.model_seed_reaction import ModelSeedReaction
from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import ROOT_DIR, DATABASE_CONFIGS


class ModelSeedReactionsDB:

    def __init__(self):

        self.__database_config = file_utilities.read_conf_file(DATABASE_CONFIGS)
        path = ROOT_DIR + self.__database_config["path_model_seed_db_reactions"]
        # scraper = ModelSeedReactionsDBScraper()
        # path = scraper()
        self.read_model_seed_reactions_database(path)

    def getReaction(self,id:str) -> ModelSeedReaction:
        return self.model_seed_reactions_database.get(id)

    def getDb(self):
        return self.model_seed_reactions_database

    def getReactionsProducingCompound(self,compound):
        reactions = self.getReactionsByCompound(compound)

        res=[]
        for reaction in reactions:
            products = reaction.getProducts()

            if compound in products:
                res.append(reaction)

        return res


    def getReactionsByCompounds(self,reactants,products):
        res=[]
        if reactants and products:
            for reaction in self.model_seed_reactions_database:
                reaction_container = self.model_seed_reactions_database[reaction]
                reaction_stoich = reaction_container.getStoichiometry()
                reaction_reactants = []
                reaction_products = []
                for comp in reaction_stoich:
                    if reaction_stoich[comp]<0:
                        reaction_reactants.append(comp)
                    else:
                        reaction_products.append(comp)

                found_reactants = 0
                found_products = 0
                for reactant in reactants:
                    if reactant in reaction_reactants:
                        found_reactants+=1
                    else:
                        break

                for product in products:
                    if product in reaction_products:
                        found_products+=1

                if found_reactants == len(reactants) and found_products == len(products):
                    res.append(reaction)

            return res


    def getReactionsByCompound(self,compound):
        res = []
        for reaction in self.model_seed_reactions_database:
            reaction_container = self.model_seed_reactions_database[reaction]
            compounds = reaction_container.getCompounds()
            if compound in compounds:
                res.append(reaction_container)
        return res

    def read_model_seed_reactions_database(self, path):
        self.model_seed_reactions_database = {}

        with open(path, "r") as reaction_file:
            reaction_file.readline()
            lines = reaction_file.readlines()
            for line in lines:
                line_lst = line.split("\t")

                is_transport = int(line_lst[5])
                is_obsolete = int(line_lst[18])
                if is_transport == 0 and is_obsolete == 0:

                    reaction_id = line_lst[0]
                    compounds_list = line_lst[16].split(";")
                    compounds=[]
                    for compound in compounds_list:
                        compound = re.sub("\[.\]","",compound)
                        compounds.append(compound)
                    if line_lst[12] != "null":
                        aliases = line_lst[12].split("|")
                        aliases_dict = {}
                        for alias in aliases:
                            clean_alias = re.sub("\s+", "", alias)
                            alias_lst = clean_alias.split(":")
                            aliases_dict[alias_lst[0]] = alias_lst[1].split(";")

                    stoichiometry = line_lst[4]
                    stoichiometry_lst = stoichiometry.split(";")
                    stoich_dict = {}
                    for compound in stoichiometry_lst:
                        if compound != "":
                            coeff_compound_id = compound.split(":")
                            coeff = float(coeff_compound_id[0])
                            compound_id = coeff_compound_id[1]
                            stoich_dict[compound_id] = coeff

                    pathways = line_lst[11]
                    pathways_dict = {}
                    if pathways != "null":
                        database_pathways_lst = pathways.split("|")

                        for database in database_pathways_lst:
                            database_name_and_path = database.split(":")
                            database_name = database_name_and_path[0].strip()
                            paths = database_name_and_path[1]
                            pathways_dict[database_name] = paths


                    equation = re.sub("\[[0-9]\]", "", line_lst[6])
                    name = line_lst[2]
                    direction = line_lst[9]
                    ec_numbers = line_lst[13].split(" | ")
                    deltag = float(line_lst[14])
                    pathways = pathways_dict
                    aliases = aliases_dict
                    stoichiometry = stoich_dict

                    # if line_lst[19] != "null":
                    #     linked_reactions = line_lst[19].split(";")
                    # else:
                    #     linked_reactions = []



                    self.model_seed_reactions_database[reaction_id] = \
                        ModelSeedReaction(reaction_id,name, equation,direction,ec_numbers, deltag,pathways,compounds,aliases,stoichiometry)




import time

from boimmgpy import definitions
from boimmgpy.database.containers.compound_node import CompoundNode
from boimmgpy.database.interfaces.boimmg_database_accessor import BOIMMGDatabaseAccessor
from boimmgpy.utilities import file_utilities
import requests

class CompoundsRestAccessor(BOIMMGDatabaseAccessor):

    def __init__(self):

        conf = file_utilities.read_conf_file(definitions.BOIMMG_DATABASE)
        self.rest_uri = conf["rest_uri"]
        self.sleep_time = 0.5

    def deserialize_compound(self,compound_dict):
        time.sleep(self.sleep_time)
        return CompoundNode(compound_dict.get("id"),compound_dict,{})


    def get_node_from_model_seed_id(self, modelseed_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/modelseed/node/" + modelseed_id)

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            node = self.deserialize_compound(results)
            return node

        return None

    def get_node_id_from_model_seed_id(self, modelseed_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/modelseed/id/" + modelseed_id)

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return None

    def get_conjugates(self, lipid_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/conjugates" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_predecessors_by_ont_id(self, lipid_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/predecessors/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_compounds_with_only_one_component(self, lipid_id,component):
        time.sleep(self.sleep_time)
        res = requests.post(self.rest_uri + "rest/compounds/same_components/" + str(lipid_id),
                            data={"components" : component})

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_compounds_with_specific_parent_within_set_of_components(self, lipid_id,components):
        time.sleep(self.sleep_time)
        res = requests.post(self.rest_uri + "rest/compounds/diff_components/" + str(lipid_id),
                            data={"components":str(components)})

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_predecessors_by_ont_id_rel_type(self, lipid_id, relationship_type):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/predecessors/" + relationship_type + "/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_all_predecessors_by_ont_id_rel_type(self, lipid_id, relationship_type):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/predecessors/all/" + relationship_type + "/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []


    def get_node_by_ont_id(self, lipid_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/compounds/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            node = self.deserialize_compound(results)
            return node

        return None

    def get_compounds_with_specific_parent_set_of_components(self, parent,components):
        time.sleep(self.sleep_time)
        res = requests.post(self.rest_uri + "rest/compounds/set_components/" + str(parent),
                            data={"components": components})

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_leaves_from_ont_id(self, lipid_id):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/leaves/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []

    def get_successors_by_ont_id_rel_type(self, lipid_id, relationship_type):
        time.sleep(self.sleep_time)
        res = requests.get(self.rest_uri + "rest/successors/" + relationship_type + "/" + str(lipid_id))

        if res.status_code == 200:
            res_dict = res.json()
            results = res_dict.get("result")
            return results

        return []


import abc

from boimmgpy.database.interfaces import node


class BOIMMGDatabaseAccessor(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'login') and
                callable(subclass.login) and
                hasattr(subclass, 'get_predecessors_by_ont_id') and
                callable(subclass.get_predecessors_by_ont_id) and
                hasattr(subclass, 'get_predecessors_by_ont_id_rel_type') and
                callable(subclass.get_predecessors_by_ont_id_rel_type) and
                hasattr(subclass, 'get_node_by_ont_id') and
                callable(subclass.get_node_by_ont_id)
                or
                NotImplemented)

    @abc.abstractmethod
    def get_predecessors_by_ont_id(self, ont_id: int) -> list:
        """Get predecessors using as parameter the database identifier"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_predecessors_by_ont_id_rel_type(self, ont_id: int, relationship_type : str) -> list:
        """Get predecessors using as parameter the database identifier and the relationship type"""

        raise NotImplementedError

    @abc.abstractmethod
    def get_node_by_ont_id(self, ont_id: int) -> node:
        """Get predecessors using as parameter the database identifier and the relationship type"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_node_from_model_seed_id(self,model_seed_id: str) -> node:
        """Get node using as parameter the model seed id"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_node_id_from_model_seed_id(self,model_seed_id: str) -> int:
        """Get database id using as parameter the model seed id"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_conjugates(self, ont_id: int) -> list:
        """Get conjugates using as parameter the database id"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_compounds_with_only_one_component(self,onto_id: int, components: list):
        """Get compounds database identifiers using as parameter a list of components"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_compounds_with_specific_parent_within_set_of_components(self,onto_id:id, components: list):
        raise NotImplementedError

    @abc.abstractmethod
    def get_all_predecessors_by_ont_id_rel_type(self, onto_id: id, relationship_type: str):
        raise NotImplementedError

    @abc.abstractmethod
    def get_compounds_with_specific_parent_set_of_components(self, onto_id: id, components: list):
        raise NotImplementedError

    @abc.abstractmethod
    def get_leaves_from_ont_id(self, ont_id: int):
        raise NotImplementedError

    @abc.abstractmethod
    def get_successors_by_ont_id_rel_type(self,onto_id: id, components: list):
        raise NotImplementedError




class IDConverter:

    def __init__(self):
        self.modelSeedToDb = {}
        self.dbToModelSeed = {}


    def convert_db_id_to_model_seed_by_db_id(self,id):
        i=0
        keys = list(self.dbToModelSeed.keys())
        latent_id =  None
        while i < len(keys):
            db = keys[i]
            ids = self.dbToModelSeed[db].keys()
            if id in ids:
                if id == "BiGG":
                    latent_id = self.dbToModelSeed[db][id]
                else:
                    return self.dbToModelSeed[db][id]

            i+=1

        if latent_id:
            return latent_id

        return None


    def convert_modelSeedId_into_other_dbID(self,modelSeedId,database_name):

        if database_name in self.modelSeedToDb.get(modelSeedId):
            return self.modelSeedToDb.get(modelSeedId).get(database_name)

        else:
            return []

    def get_all_aliases_by_modelSeedID(self,modelSeedId):
        if modelSeedId in self.modelSeedToDb.keys():
            res = self.modelSeedToDb[modelSeedId]
            if "BiGG" in res and "BiGG1" in res:
                del res["BiGG"]
                bigg_ids = res["BiGG1"]
                res["BiGG"] = bigg_ids


            res["ModelSEED"] = [modelSeedId]

            res = self.__prune_dictionary(res)

            return res
        else:
            return {}

    def __prune_dictionary(self,dictionary):
        res = dictionary.copy()
        for key in dictionary:
            if "" in dictionary[key]:
                del res[key]
        return res

    def convert_dbID_into_modelSeedId(self,database_name,id):
        if id in self.dbToModelSeed.get(database_name).keys():
            return self.dbToModelSeed.get(database_name).get(id)
        else:
            return None

    def get_modelSeedIdToDb(self):
        return self.modelSeedToDb

    def get_dbIdToModelSeed(self):
        return self.dbToModelSeed


import re

from boimmgpy.kegg.kegg_entity import KeggEntity


class KeggReaction(KeggEntity):

    def __init__(self,entry, byModules=True):
        super().__init__(entry)
        self.byModules = byModules
        self.__get_all_information()

    def __get_all_information(self):
        info = self.get_raw_data()
        rclass_go = False

        for text in info.split("\n"):
            new_line = re.sub('\s{2,}', "\t", text)

            if re.search("\ANAME", new_line):
                self.name = new_line.split("\t")[1]

            elif re.search("\AEQUATION", new_line):
                self.equation = new_line.split("\t")[1]
                self.get_reactants_and_products()

            elif re.search("\AENZYME", new_line):
                rclass_go = False
                self.enzymes=new_line.split("\t")[1:]

            elif re.search("\APATHWAY", new_line):
                rclass_go = False

            elif re.search("\ARCLASS", new_line):
                rclass_go=True
                self.rclass = {}
                compounds = new_line.split("\t")[2]
                compound_pair = compounds.split("_")
                compound1 = compound_pair[0]
                compound2 = compound_pair[1]
                self.rclass[compound1] = compound2
                self.rclass[compound2] = compound1


            # elif rclass_go:
            #     compounds = new_line.split("\t")[2]
            #     compound_pair = compounds.split("_")
            #     compound1 = compound_pair[0]
            #     compound2 = compound_pair[1]
            #     self.rclass[compound1] = compound2
            #     self.rclass[compound2] = compound1

    def get_name(self):
        try:
            name = self.name
            return name
        except:
            return None



    def get_reactants_and_products(self):
        # equation = re.sub("\s","",self.equation)
        match = re.search("<*=>*",self.equation)
        start = match.start()
        end = match.end()

        self.direction = self.equation[start:end]
        if self.direction == "<=>":
            self.reversibility = True
        else:
            self.reversibility =False

        sides = re.split("<*=>*",self.equation)
        reactants = sides[0].split("+")
        self.reactants={}
        self.products={}

        for reactant in reactants:
            coef = 1
            r = reactant.split(" ")
            for l in r:
                if re.search("^[0-9][0-9]*$",l):
                    coef = int(re.sub("\s+","",l))
                elif re.search("C[0-9]*",l):
                    compound = re.sub("\s+","",l)
            self.reactants[compound] = coef * -1

        products = sides[1].split("+")
        for product in products:
            r = product.split(" ")
            coef = 1
            for l in r:
                if re.search("^[0-9][0-9]*$", l):
                    coef = int(re.sub("\s+", "", l))
                elif re.search("C[0-9]*",l):
                    compound = re.sub("\s+", "", l)

            self.products[compound] = coef

        self.compounds = {}
        self.compounds.update(self.products)
        self.compounds.update(self.reactants)

    def get_id(self):
        return self.id

    def get_stoichiometry(self):
        return self.compounds

    def get_compounds(self):
        return self.compounds

    def get_reactants(self):
        return self.reactants

    def get_products(self):
        return self.products

    def get_reversibility(self):
        return self.reversibility

    def get_direction(self):
        return self.direction

    def get_equation(self):
        return self.equation

    def get_rclass(self):
        return self.rclass


def is_in_tuple_list(tl, val):
    res = False
    for (x, y) in tl:
        if val == x:    return True
    return res


class Graph:

    def __init__(self, g={}):
        self.graph = g

    def delete_node(self,node):
        predecessors = self.get_predecessors(node)
        for pre in predecessors:
            self.graph[pre].remove(node)
        del self.graph[node]

    def print_graph(self):
        for v in self.graph.keys():
            print(v, "->", self.graph[v])

    def get_nodes(self):
        return list(self.graph.keys())

    def get_edges(self):
        edges = []
        for v in self.graph.keys():
            for d in self.graph[v]:
                edges.append((v, d))
        return edges

    def add_vertex(self, v):
        if v not in self.graph:
            self.graph[v] = []

    def add_edge(self, o, d):
        if o not in self.graph:
            self.add_vertex(o)

        if d not in self.graph:
            self.add_vertex(d)

        if d not in self.graph[o]:
            self.graph[o].append(d)

    def get_successors(self, v):
        return self.graph[v]

    def get_predecessors(self, v):
        lst = []
        for pre in self.graph:
            if v in self.graph[pre]:
                lst.append(pre)
        return sorted(lst)

    def get_adjacents(self, v):
        lst_suc = self.get_successors(v)
        lst_pre = self.get_predecessors(v)
        res = lst_pre
        for v1 in lst_suc:
            if v1 not in res:
                res.append(v1)
        return res

    def out_degree(self, v):
        return len(self.graph[v])

    def in_degree(self, v):
        return len(self.get_predecessors(v))

    def degree(self, v):
        return len(self.get_adjacents(v))

    def reachable_bfs(self, v):
        l = [v]
        res = [v]
        dist = 0
        while len(l) > 0:
            node = l.pop(0)
            if node != v: res.append(node)
            for elem in self.graph[node]:
                if elem not in res and elem not in l and elem != node:
                    l.append(elem)
        return res

    def reachable_dfs(self, v):
        l = [v]
        res = []
        while len(l) > 0:
            node = l.pop(0)
            if node != v: res.append(node)
            for elem in self.graph[node]:
                if elem not in res and elem not in l:
                    l.insert(0, elem)
        return res

    def distance(self, s, d):
        if s == d:
            return 0
        l = [(s, 0)]
        visited = [s]
        while len(l) > 0:
            node, dist = l.pop(0)
            for elem in self.graph[node]:
                if elem == d:
                    return dist + 1
                elif elem not in visited:
                    l.append((elem, dist + 1))
                    visited.append(elem)
        return None

    def reachable_with_dist(self, s):
        res = []
        l = [(s, 0)]
        while len(l) > 0:
            node, dist = l.pop(0)
            if node != s:    res.append((node, dist))
            for elem in self.graph[node]:
                if not is_in_tuple_list(l, elem) and not is_in_tuple_list(res, elem):
                    l.append((elem, dist + 1))
        return res

    def shortest_path(self, s, d):
        if s == d:
            return []
        l = [(s, [])]
        visited = [s]
        while len(l) > 0:
            node, preds = l.pop(0)
            for elem in self.graph[node]:
                if elem == d:
                    return preds + [node, elem]
                elif elem not in visited:
                    l.append((elem, preds + [node]))
                    visited.append(elem)
        return None

    def node_has_cycle(self, v):
        l = [v]
        res = False
        visited = [v]
        while len(l) > 0:
            node = l.pop(0)
            for elem in self.graph[node]:
                if elem == v:
                    return True
                elif elem not in visited:
                    l.append(elem)
                    visited.append(elem)
        return res

    def has_cycle(self):
        res = False
        for v in self.graph.keys():
            if self.node_has_cycle(v):    return True
        return res

    def all_degrees(self, deg_type="inout"):
        degs = {}
        for v in self.graph.keys():
            if deg_type == "out" or deg_type == "inout":
                degs[v] = len(self.graph[v])
            else:
                degs[v] = 0
        if deg_type == "in" or deg_type == "inout":
            for v in self.graph.keys():
                for d in self.graph[v]:
                    if deg_type == "in" or v not in self.graph[d]:
                        degs[d] = degs[d] + 1
        return degs

    def mean_degree(self, deg_type="inout"):
        degs = self.all_degrees(deg_type)
        return sum(degs.values()) / float(len(degs))

    def prob_degree(self, deg_type="inout"):
        degs = self.all_degrees(deg_type)
        res = {}
        for k in degs.keys():
            if degs[k] in res.keys():
                res[degs[k]] += 1
            else:
                res[degs[k]] = 1
        for k in res.keys():
            res[k] /= float(len(degs))
        return res

    def mean_distance(self):
        tot = 0
        num_reachable = 0
        for k in self.graph.keys():
            distk = self.reachable_with_dist(k)
            for _, dist in distk:
                tot += dist
            num_reachable += len(distk)
        meandist = float(tot) / num_reachable
        n = len(self.getNodes())
        return meandist, float(num_reachable) / ((n - 1) * n)

    def clustering_coef(self, v):
        adjs = self.get_adjacents(v)
        if len(adjs) <= 1:    return 0.0
        ligs = 0
        for i in adjs:
            for j in adjs:
                if i != j:
                    if j in self.graph[i] or i in self.graph[j]:
                        ligs = ligs + 1
        return float(ligs) / (len(adjs) * (len(adjs) - 1))

    def all_clustering_coefs(self):
        ccs = {}
        for k in self.graph.keys():
            ccs[k] = self.clustering_coef(k)
        return ccs

    def mean_clustering_coef(self):
        ccs = self.all_clustering_coefs()
        return sum(ccs.values()) / float(len(ccs))

    def mean_clustering_perdegree(self, deg_type="inout"):
        degs = self.all_degrees(deg_type)
        ccs = self.all_clustering_coefs()
        degs_k = {}
        for k in degs.keys():
            if degs[k] in degs_k.keys():
                degs_k[degs[k]].append(k)
            else:
                degs_k[degs[k]] = [k]
        ck = {}
        for k in degs_k.keys():
            tot = 0
            for v in degs_k[k]:    tot += ccs[v]
            ck[k] = float(tot) / len(degs_k[k])
        return ck

    def check_if_valid_path(self, p):
        if p[0] not in self.graph.keys():    return False
        for i in range(1, len(p)):
            if p[i] not in self.graph.keys() or p[i] not in self.graph[p[i - 1]]:
                return False
        return True

    def check_if_hamiltonian_path(self, p):
        if not self.check_if_valid_path(p):
            return False
        to_visit = list(self.get_nodes())
        if len(to_visit) != len(to_visit):
            return False

        for i in range(len(p)):
            if p[i] in to_visit:
                to_visit.remove(p[i])
            else:
                return False

        if not to_visit:
            return True
        else:
            return False

    def check_balanced_node(self, node):
        return self.in_degree(node) == self.out_degree(node)

    def check_balanced_graph(self):
        for n in self.graph.keys():
            if not self.check_balanced_node(n):    return False
        return True

    def eulerian_cycle(self):
        if not self.check_balanced_graph():
            return None
        edges_visit = list(self.get_edges())
        res = []
        while edges_visit:
            pair = edges_visit[0]
            i = 1
            if res != []:
                while pair[0] not in res:
                    pair = edges_visit[i]
                    i = i + 1
            edges_visit.remove(pair)
            start, nxt = pair
            cycle = [start, nxt]
            while nxt != start:
                for suc in self.graph[nxt]:
                    if (nxt, suc) in edges_visit:
                        print(pair)
                        pair = (nxt, suc)
                        nxt = suc
                        cycle.append(nxt)
                        print(cycle)
                        edges_visit.remove(pair)
            if not res:
                res = cycle
            else:
                pos = res.index(cycle[0])
                print("res: ", res, pos)
                for i in range(len(cycle) - 1): res.insert(pos + i + 1, cycle[i + 1])
                print("res: ", res)
        return res

    def check_nearly_balanced_graph(self):
        res = None, None
        for n in self.graph.keys():
            indeg = self.in_degree(n)
            outdeg = self.out_degree(n)
            if indeg - outdeg == 1 and res[1] is None:
                res = res[0], n
            elif indeg - outdeg == -1 and res[0] is None:
                res = n, res[1]
            elif indeg == outdeg:
                pass
            else:
                return None, None
        return res

    def eulerian_path(self):
        unb = self.check_nearly_balanced_graph()
        if unb[0] is None or unb[1] is None: return None
        self.graph[unb[1]].append(unb[0])
        cycle = self.eulerian_cycle()
        for i in range(len(cycle) - 1):
            if cycle[i] == unb[1] and cycle[i + 1] == unb[0]:
                break
        path = cycle[i + 1:] + cycle[1:i + 1]
        return path

    def closeness_centrality(self, node):
        dist = self.reachable_with_dist(node)
        if len(dist) == 0: return 0.0
        s = 0.0
        for d in dist: s += d[1]
        return len(dist) / s

    def highest_closeness(self, top=10):
        cc = {}
        for k in self.graph.keys(): cc[k] = self.closeness_centrality(k)
        ord_cl = sorted(list(cc.items()), key=lambda x: x[1], reverse=True)
        return list(map(lambda x: x[0], ord_cl[:top]))

    def betweenness_centrality(self, node):
        total_sp = 0
        sps_with_node = 0
        for s in self.graph.keys():
            for t in self.graph.keys():
                if s != t and s != node and t != node:
                    sp = self.shortest_path(s, t)
                    if sp is not None:
                        total_sp += 1
                        if node in sp: sps_with_node += 1
        return sps_with_node / total_sp
























